# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Actuation of joints."""

from typing import Tuple

from flax import struct
import jax
import jax.numpy as jnp

from brax.physics import config_pb2
from brax.physics import joints
from brax.physics.base import P, QP, take


@struct.dataclass
class Actuator:
  """Applies a torque to a joint."""
  joint: joints.Joint
  strength: jnp.ndarray
  act_idx: jnp.ndarray
  config: config_pb2.Config = struct.field(pytree_node=False)

  @classmethod
  def from_config(cls, config: config_pb2.Config,
                  joint: joints.Joint) -> "Actuator":
    """Creates an actuator from a config."""
    joint_type = type(joint).__name__.lower()
    act_type = cls.__name__.lower()
    joints_filtered = [
        j for j in config.joints if joints.type_to_dof[joint_type] ==
        joints.lim_to_dof[len(j.angle_limit)]
    ]
    joint_map = {j.name: i for i, j in enumerate(joints_filtered)}
    actuators = [
        a for a in config.actuators
        if a.HasField(act_type) and a.joint in joint_map
    ]
    if not actuators:
      return cls(*[None] * 4)
    joint_idx = jnp.array([joint_map[a.joint] for a in actuators])
    joint = take(joint, joint_idx)  # ensure joints are in the right order
    strength = jnp.array([a.strength for a in actuators])
    act_idx = jnp.array([_act_idx(config, a.name) for a in actuators])
    return cls(joint, strength, act_idx, config)

  @jax.vmap
  def _apply(self, target: jnp.ndarray, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    raise NotImplementedError()  # child must override

  def apply(self, qp: QP, target: jnp.ndarray) -> P:
    """Applies torque to satisfy a target angle on a revolute joint.

    Args:
      qp: State data for system
      target: User-defined target angle for the actuator in degrees

    Returns:
      dP: The impulses that drive a parent and child towards a target angle
    """
    if not self.config:
      return P(jnp.zeros_like(qp.vel), jnp.zeros_like(qp.ang))

    qp_p = take(qp, self.joint.body_p.idx)
    qp_c = take(qp, self.joint.body_c.idx)
    target = take(target, self.act_idx)
    dang_p, dang_c = self._apply(target, qp_p, qp_c)

    # sum together all impulse contributions across parents and children
    body_idx = jnp.concatenate((self.joint.body_p.idx, self.joint.body_c.idx))
    dp_ang = jnp.concatenate((dang_p, dang_c))
    dp_ang = jax.ops.segment_sum(dp_ang, body_idx, len(self.config.bodies))

    return P(vel=jnp.zeros_like(qp.vel), ang=dp_ang)


@struct.dataclass
class Angle(Actuator):
  """Applies torque to satisfy a target angle of a joint."""

  @jax.vmap
  def _apply(self, target: jnp.ndarray, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    axis, angle = self.joint.axis_angle(qp_p, qp_c)
    axis, angle = jnp.array(axis), jnp.array(angle)

    # torque grows as target angle diverges from current angle
    target *= jnp.pi / 180.
    target = jnp.clip(target, self.joint.limit[:, 0], self.joint.limit[:, 1])
    torque = (target - angle) * self.strength
    torque = jnp.sum(jax.vmap(jnp.multiply)(axis, torque), axis=0)

    dang_p = jnp.matmul(self.joint.body_p.inertia, -torque)
    dang_c = jnp.matmul(self.joint.body_c.inertia, torque)

    return dang_p, dang_c


@struct.dataclass
class Torque(Actuator):
  """Applies a target torque to a joint."""

  @jax.vmap
  def _apply(self, target: jnp.ndarray, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    axis, angle = self.joint.axis_angle(qp_p, qp_c)
    axis, angle = jnp.array(axis), jnp.array(angle)

    # clip torque if outside joint angle limits
    torque = target * self.strength
    torque = jnp.where(angle < self.joint.limit[:, 0], 0, torque)
    torque = jnp.where(angle > self.joint.limit[:, 1], 0, torque)
    torque = jnp.sum(jax.vmap(jnp.multiply)(axis, torque), axis=0)

    dang_p = jnp.matmul(self.joint.body_p.inertia, torque)
    dang_c = jnp.matmul(self.joint.body_c.inertia, -torque)

    return dang_p, dang_c


def _find_joint(config, name):
  for j in config.joints:
    if j.name == name:
      return j
  raise ValueError(f"Joint [{name}] does not exist.")


def _act_idx(config, name):
  """Returns the indices that an actuator occupies in an array."""
  current_idx = 0
  for actuator in config.actuators:
    joint = _find_joint(config, actuator.joint)
    dof = len(joint.angle_limit)
    if actuator.name == name:
      if dof == 1:
        return current_idx
      else:
        return tuple(range(current_idx, current_idx + dof))
    current_idx += dof
  raise ValueError(f"Actuator [{name}] does not exist.")



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Calculate collisions of bodies."""

import itertools
from typing import List, Tuple

from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics import math
from brax.physics.base import P, QP, euler_to_quat, take, vec_to_np
from flax import struct
import jax
from jax import ops
import jax.numpy as jnp


class BoxPlane:
  """A collision between a box corner and a static plane."""

  def __init__(self, config: config_pb2.Config):
    self.config = config
    self.pairs = _find_body_pairs(config, 'box', 'plane')
    if not self.pairs:
      return

    body_idx = {b.name: i for i, b in enumerate(config.bodies)}
    box_idxs = []
    plane_idxs = []
    corners = []

    for box, plane in self.pairs:
      if not plane.frozen.all:
        raise ValueError('active planes unsupported: %s' % plane)
      for i in range(8):
        box_idxs.append(body_idx[box.name])
        plane_idxs.append(body_idx[plane.name])
        corner = jnp.array(
            [i % 2 * 2 - 1, 2 * (i // 4) - 1, i // 2 % 2 * 2 - 1],
            dtype=jnp.float32)
        col = box.colliders[0]
        corner = corner * vec_to_np(col.box.halfsize)
        corner = math.rotate(corner, euler_to_quat(col.rotation))
        corner = corner + vec_to_np(col.position)
        corners.append(corner)

    body = bodies.Body.from_config(config)
    self.box = take(body, jnp.array(box_idxs))
    self.plane = take(body, jnp.array(plane_idxs))
    self.corner = jnp.array(corners)

  def apply(self, qp: QP, dt: float) -> P:
    """Returns impulse from a collision between box corners and a static plane.

    Note that impulses returned by this function are *not* scaled by dt when
    applied to parts.  Collision impulses are applied directly as velocity and
    angular velocity updates.

    Args:
      qp: Coordinate/velocity frame of the bodies.
      dt: Integration time step length.

    Returns:
      dP: Delta velocity to apply to the box bodies in the collision.
      colliding: Mask for each body: 1 = colliding, 0 = not colliding.
    """
    if not self.pairs:
      return P(jnp.zeros_like(qp.vel), jnp.zeros_like(qp.ang))

    @jax.vmap
    def apply(box, corner, qp_box, qp_plane):
      pos, vel = math.to_world(qp_box, corner)
      normal = math.rotate(jnp.array([0.0, 0.0, 1.0]), qp_plane.rot)
      penetration = jnp.dot(pos - qp_plane.pos, normal)
      dp = _collide(self.config, box, qp_box, pos, vel, normal, penetration, dt)
      collided = jnp.where(penetration < 0., 1., 0.)
      return dp, collided

    qp_box = take(qp, self.box.idx)
    qp_plane = take(qp, self.plane.idx)
    dp, colliding = apply(self.box, self.corner, qp_box, qp_plane)

    # collapse/sum across all corners
    num_bodies = len(self.config.bodies)
    colliding = ops.segment_sum(colliding, self.box.idx, num_bodies)
    vel = ops.segment_sum(dp.vel, self.box.idx, num_bodies)
    ang = ops.segment_sum(dp.ang, self.box.idx, num_bodies)

    # equally distribute contact force over each box
    vel = vel / jnp.reshape(1e-8 + colliding, (vel.shape[0], 1))
    ang = ang / jnp.reshape(1e-8 + colliding, (ang.shape[0], 1))

    return P(vel, ang)


class BoxHeightMap:
  """A collision between box corners and a height map."""

  def __init__(self, config: config_pb2.Config):
    self.config = config
    self.pairs = _find_body_pairs(config, 'box', 'heightMap')
    if not self.pairs:
      return

    body_idx = {b.name: i for i, b in enumerate(config.bodies)}
    box_idxs = []
    height_map_idxs = []
    corners = []
    heights = []
    sizes = []
    mesh_sizes = []

    for box, height_map in self.pairs:
      if not height_map.frozen.all:
        raise ValueError('active height maps unsupported: %s' % height_map)
      for i in range(8):
        box_idxs.append(body_idx[box.name])
        height_map_idxs.append(body_idx[height_map.name])
        corner = jnp.array(
            [i % 2 * 2 - 1, 2 * (i // 4) - 1, i // 2 % 2 * 2 - 1],
            dtype=jnp.float32)
        col = box.colliders[0]
        corner = corner * vec_to_np(col.box.halfsize)
        corner = math.rotate(corner, euler_to_quat(col.rotation))
        corner = corner + vec_to_np(col.position)
        corners.append(corner)

        mesh_size = int(
            jnp.round(jnp.sqrt(len(height_map.colliders[0].heightMap.data))))
        if len(height_map.colliders[0].heightMap.data) != mesh_size**2:
          raise ValueError(
              'data length for an height map should be a perfect square.')

        height = jnp.array(height_map.colliders[0].heightMap.data).reshape(
            (mesh_size, mesh_size))
        heights.append(height)
        sizes.append(height_map.colliders[0].heightMap.size)
        mesh_sizes.append(mesh_size)

    body = bodies.Body.from_config(config)
    self.box = take(body, jnp.array(box_idxs))
    self.height_map = take(body, jnp.array(height_map_idxs))
    self.corner = jnp.array(corners)
    self.size = jnp.array(sizes)
    self.mesh_size = jnp.array(mesh_sizes)
    self.heights = jnp.array(heights)

  def apply(self, qp: QP, dt: float) -> P:
    """Returns impulse from a collision between box corners and a static height map.

    Note that impulses returned by this function are *not* scaled by dt when
    applied to parts.  Collision impulses are applied directly as velocity and
    angular velocity updates.
    Args:
      qp: Coordinate/velocity frame of the bodies.
      dt: Integration time step length.

    Returns:
      dP: Delta velocity to apply to the box bodies in the collision.
      colliding: Mask for each body: 1 = colliding, 0 = not colliding.
    """
    if not self.pairs:
      return P(jnp.zeros_like(qp.vel), jnp.zeros_like(qp.ang))

    @jax.vmap
    def apply(box, corner, qp_box, qp_height_map, size, mesh_size, heights):
      world_pos, vel = math.to_world(qp_box, corner)
      pos = math.inv_rotate(world_pos - qp_height_map.pos, qp_height_map.rot)
      uv_pos = (pos[:2]) / size * (mesh_size - 1)

      # find the square in the mesh that enclose the candidate point, with mesh
      # indices ux_idx, ux_udx_u, uv_idx_v, uv_idx_uv.
      uv_idx = jnp.floor(uv_pos).astype(jnp.int32)
      uv_idx_u = uv_idx + jnp.array([1, 0], dtype=jnp.int32)
      uv_idx_v = uv_idx + jnp.array([0, 1], dtype=jnp.int32)
      uv_idx_uv = uv_idx + jnp.array([1, 1], dtype=jnp.int32)

      # find the orientation of the triangle of this square that encloses the
      # candidate point
      delta_uv = uv_pos - uv_idx
      mu = jnp.where(
          delta_uv[0] + delta_uv[1] < 1, 1,
          -1)  # whether the corner lies on the first or secound triangle

      # compute the mesh indices of the vertices of this triangle
      point_0 = jnp.where(delta_uv[0] + delta_uv[1] < 1, uv_idx, uv_idx_uv)
      point_1 = jnp.where(delta_uv[0] + delta_uv[1] < 1, uv_idx_u, uv_idx_v)
      point_2 = jnp.where(delta_uv[0] + delta_uv[1] < 1, uv_idx_v, uv_idx_u)

      h0 = heights[point_0[0], point_0[1]]
      h1 = heights[point_1[0], point_1[1]]
      h2 = heights[point_2[0], point_2[1]]

      raw_normal = jnp.array(
          [-mu * (h1 - h0), -mu * (h2 - h0), 1 * (size / (mesh_size - 1))])
      normal = raw_normal / jnp.linalg.norm(raw_normal)
      rotated_normal = math.rotate(normal, qp_height_map.rot)

      pos_0 = jnp.array([
          point_0[0] * size / (mesh_size - 1),
          point_0[1] * size / (mesh_size - 1), h0
      ])
      penetration = jnp.dot(pos - pos_0, normal)

      dp = _collide(self.config, box, qp_box, pos, vel, rotated_normal,
                    penetration, dt)
      collided = jnp.where(penetration < 0., 1., 0.)
      return dp, collided

    qp_box = take(qp, self.box.idx)
    qp_height_map = take(qp, self.height_map.idx)
    dp, colliding = apply(self.box, self.corner, qp_box, qp_height_map,
                          self.size, self.mesh_size, self.heights)

    # collapse/sum across all corners
    num_bodies = len(self.config.bodies)
    colliding = ops.segment_sum(colliding, self.box.idx, num_bodies)
    vel = ops.segment_sum(dp.vel, self.box.idx, num_bodies)
    ang = ops.segment_sum(dp.ang, self.box.idx, num_bodies)

    # equally distribute contact force over each box (corner ?)
    vel = vel / jnp.reshape(1e-8 + colliding, (vel.shape[0], 1))
    ang = ang / jnp.reshape(1e-8 + colliding, (ang.shape[0], 1))

    return P(vel, ang)


class CapsulePlane:
  """A collision between a capsule and a static plane."""

  def __init__(self, config: config_pb2.Config):
    self.config = config
    self.pairs = _find_body_pairs(config, 'capsule', 'plane')
    self.pairs += _find_body_pairs(config, 'sphere', 'plane')
    if not self.pairs:
      return

    body_idx = {b.name: i for i, b in enumerate(config.bodies)}
    for _, plane in self.pairs:
      if not plane.frozen.all:
        raise ValueError('active planes unsupported: %s' % plane)

    cap_idx = []
    plane_idx = []
    cap_end = []
    cap_radius = []
    for cap, plane in self.pairs:
      if cap.colliders[0].WhichOneof('type') == 'sphere':
        capsule = config_pb2.Collider.Capsule()
        capsule.radius = cap.colliders[0].sphere.radius
        capsule.length = 2 * capsule.radius
        capsule.end = 1
      else:
        capsule = cap.colliders[0].capsule
      ends = [capsule.end] if capsule.end else [-1, 1]
      for end in ends:
        cap_idx.append(body_idx[cap.name])
        plane_idx.append(body_idx[plane.name])
        cap_pos = vec_to_np(cap.colliders[0].position)
        cap_rot = euler_to_quat(cap.colliders[0].rotation)
        cap_axis = math.rotate(jnp.array([0., 0., 1.]), cap_rot)
        cap_arm = capsule.length / 2 - capsule.radius
        cap_end.append(cap_pos + end * cap_axis * cap_arm)
        cap_radius.append(capsule.radius)

    body = bodies.Body.from_config(config)
    self.cap = take(body, jnp.array(cap_idx))
    self.plane = take(body, jnp.array(plane_idx))
    self.cap_end = jnp.array(cap_end)
    self.cap_radius = jnp.array(cap_radius)

  def apply(self, qp: QP, dt: float) -> P:
    """Returns an impulse from a collision between a capsule and a static plane.

    Note that impulses returned by this function are *not* scaled by dt when
    applied to parts.  Collision impulses are applied directly as velocity and
    angular velocity updates.

    Args:
      qp: Coordinate/velocity frame of the bodies.
      dt: Integration time step length.

    Returns:
      dP: Delta velocity to apply to the capsule bodies in the collision.
      colliding: Mask for each body: 1 = colliding, 0 = not colliding.
    """
    if not self.pairs:
      return P(jnp.zeros_like(qp.vel), jnp.zeros_like(qp.ang))

    @jax.vmap
    def apply(cap, cap_end, radius, qp_cap, qp_plane):
      cap_end_world = qp_cap.pos + math.rotate(cap_end, qp_cap.rot)
      normal = math.rotate(jnp.array([0.0, 0.0, 1.0]), qp_plane.rot)
      pos = cap_end_world - normal * radius
      rpos_off = pos - qp_cap.pos
      rvel = jnp.cross(qp_cap.ang, rpos_off)
      vel = qp_cap.vel + rvel

      penetration = jnp.dot(pos - qp_plane.pos, normal)
      dp = _collide(self.config, cap, qp_cap, pos, vel, normal, penetration, dt)
      colliding = jnp.where(penetration < 0., 1., 0.)
      return dp, colliding

    qp_cap = take(qp, self.cap.idx)
    qp_plane = take(qp, self.plane.idx)
    dp, colliding = apply(self.cap, self.cap_end, self.cap_radius, qp_cap,
                          qp_plane)

    # sum across both contact points
    num_bodies = len(self.config.bodies)
    colliding = ops.segment_sum(colliding, self.cap.idx, num_bodies)
    vel = ops.segment_sum(dp.vel, self.cap.idx, num_bodies)
    ang = ops.segment_sum(dp.ang, self.cap.idx, num_bodies)

    # equally distribute contact force over possible collision points
    vel = vel / jnp.reshape(1e-8 + colliding, (vel.shape[0], 1))
    ang = ang / jnp.reshape(1e-8 + colliding, (ang.shape[0], 1))

    return P(vel, ang)


class CapsuleCapsule:
  """A collision between two capsules."""

  @struct.dataclass
  class Capsule:
    body: bodies.Body
    radius: jnp.ndarray
    end: jnp.ndarray

  def __init__(self, config: config_pb2.Config):
    self.config = config
    self.pairs = _find_body_pairs(config, 'capsule', 'capsule')
    if not self.pairs:
      return None

    body_idx = {b.name: i for i, b in enumerate(config.bodies)}
    body_idxs = []
    radii = []
    ends = []

    for body_a, body_b in self.pairs:
      col_a, col_b = body_a.colliders[0], body_b.colliders[0]
      body_idxs.append((body_idx[body_a.name], body_idx[body_b.name]))
      radii.append((col_a.capsule.radius, col_b.capsule.radius))
      def cap_end(col):
        axis = math.rotate(jnp.array([0., 0., 1.]), euler_to_quat(col.rotation))
        segment_length = col.capsule.length / 2. - col.capsule.radius
        return vec_to_np(col.position) + axis * segment_length
      ends.append((cap_end(col_a), cap_end(col_b)))

    body = bodies.Body.from_config(config)
    self.cap_a = CapsuleCapsule.Capsule(
        body=take(body, jnp.array([a for a, _ in body_idxs])),
        radius=jnp.array([a for a, _ in radii]),
        end=jnp.array([a for a, _ in ends]))
    self.cap_b = CapsuleCapsule.Capsule(
        body=take(body, jnp.array([b for _, b in body_idxs])),
        radius=jnp.array([b for _, b in radii]),
        end=jnp.array([b for _, b in ends]))

  def apply(self, qp: QP, dt: float) -> P:
    """Returns impulses between capsules in collision.

    Args:
      qp: Coordinate/velocity frame of the bodies.
      dt: Integration time step length.

    Returns:
      dP: Delta velocity to apply to the capsule bodies in the collision.
      colliding: Mask for each body: 1 = colliding, 0 = not colliding.
    """
    if not self.pairs:
      return P(jnp.zeros_like(qp.vel), jnp.zeros_like(qp.ang))

    @jax.vmap
    def apply(cap_a, cap_b, qp_a, qp_b):
      """Extracts collision points and applies collision to capsules."""

      def endpoints(cap, qp):
        end = math.rotate(cap.end, qp.rot)
        return qp.pos + end, qp.pos - end

      def closest_segment_point(a, b, pt):
        ab = b - a
        t = jnp.dot(pt - a, ab) / jnp.dot(ab, ab)
        return a + jnp.clip(t, 0., 1.) * ab

      a_A, a_B = endpoints(cap_a, qp_a)
      b_A, b_B = endpoints(cap_b, qp_b)

      v0 = b_A - a_A
      v1 = b_B - a_A
      v2 = b_A - a_B
      v3 = b_B - a_B

      d0 = jnp.dot(v0, v0)
      d1 = jnp.dot(v1, v1)
      d2 = jnp.dot(v2, v2)
      d3 = jnp.dot(v3, v3)

      bestA = jnp.where((d2 < d0) | (d2 < d1) | (d3 < d0) | (d3 < d1), a_B, a_A)
      bestB = closest_segment_point(b_A, b_B, bestA)
      bestA = closest_segment_point(a_A, a_B, bestB)

      penetration_vec = bestB - bestA
      dist = math.safe_norm(penetration_vec)
      collision_normal = penetration_vec / (1e-6 + dist)
      penetration = dist - cap_a.radius - cap_b.radius
      pos_c = (bestA + bestB) / 2.

      dp_a, dp_b = _collide_pair(self.config, cap_a.body, cap_b.body, qp_a,
                                 qp_b, pos_c, collision_normal, penetration, dt)
      return dp_a, dp_b

    qp_a = take(qp, self.cap_a.body.idx)
    qp_b = take(qp, self.cap_b.body.idx)
    dp_a, dp_b = apply(self.cap_a, self.cap_b, qp_a, qp_b)

    num_bodies = len(self.config.bodies)
    vel_a = ops.segment_sum(dp_a.vel, self.cap_a.body.idx, num_bodies)
    ang_a = ops.segment_sum(dp_a.ang, self.cap_a.body.idx, num_bodies)
    vel_b = ops.segment_sum(dp_b.vel, self.cap_b.body.idx, num_bodies)
    ang_b = ops.segment_sum(dp_b.ang, self.cap_b.body.idx, num_bodies)

    return P(vel_a + vel_b, ang_a + ang_b)


def _find_body_pairs(
    config: config_pb2.Config, type_a: str,
    type_b: str) -> List[Tuple[config_pb2.Body, config_pb2.Body]]:
  """Returns body pairs with colliders matching types."""
  include = {(f.first, f.second) for f in config.collide_include}
  include |= {(f.second, f.first) for f in config.collide_include}

  # exclude collisions between parents and children of joints, unless
  # explicitly included
  ignore = {(j.parent, j.child) for j in config.joints}
  ignore |= {(j.child, j.parent) for j in config.joints}
  ignore -= include

  # flatten and emit one collider per body
  flat_bodies = []
  for body in config.bodies:
    for collider in body.colliders:
      new_body = config_pb2.Body()
      new_body.CopyFrom(body)
      new_body.ClearField('colliders')
      new_body.colliders.append(collider)
      flat_bodies.append(new_body)

  ret = []
  for body_a, body_b in itertools.combinations(flat_bodies, 2):
    if include and ((body_a.name, body_b.name) not in include):
      continue
    if (body_a.name, body_b.name) in ignore:
      continue
    if body_a.frozen.all and body_b.frozen.all:
      continue

    collider_type_a = body_a.colliders[0].WhichOneof('type')
    collider_type_b = body_b.colliders[0].WhichOneof('type')
    if (collider_type_a, collider_type_b) == (type_a, type_b):
      ret.append((body_a, body_b))
    elif (collider_type_b, collider_type_a) == (type_a, type_b):
      ret.append((body_b, body_a))
  return ret


def _collide(config: config_pb2.Config, body: bodies.Body, qp: QP,
             pos_c: jnp.ndarray, vel_c: jnp.ndarray, normal_c: jnp.ndarray,
             penetration: float, dt: float) -> P:
  """Calculates velocity change due to a collision.

  Args:
    config: A brax config.
    body: Body participating in collision
    qp: State for body
    pos_c: Where the collision is occuring in world space
    vel_c: How fast the collision is happening
    normal_c: Normal vector of surface providing collision
    penetration: Amount of penetration between part and plane
    dt: Integration timestep length

  Returns:
    dP: The impulse on this body result from the collision.
  """
  rel_pos_a = pos_c - qp.pos

  baumgarte_rel_vel = config.baumgarte_erp * penetration / dt
  normal_rel_vel = jnp.dot(normal_c, vel_c)

  temp1 = jnp.matmul(body.inertia, jnp.cross(rel_pos_a, normal_c))
  ang = jnp.dot(normal_c, jnp.cross(temp1, rel_pos_a))
  impulse = (-1. *
             (1. + config.elasticity) * normal_rel_vel - baumgarte_rel_vel) / (
                 (1. / body.mass) + ang)

  dp_n = body.impulse(qp, impulse * normal_c, pos_c)

  # apply drag due to friction acting parallel to the surface contact
  rel_vel_d = vel_c - normal_rel_vel * normal_c
  impulse_d = math.safe_norm(rel_vel_d) / ((1. / (body.mass)) + ang)
  # drag magnitude cannot exceed max friction
  impulse_d = jnp.where(impulse_d < config.friction * impulse, impulse_d,
                        config.friction * impulse)
  dir_d = rel_vel_d / (1e-6 + math.safe_norm(rel_vel_d))
  dp_d = body.impulse(qp, -impulse_d * dir_d, pos_c)

  # apply collision normal if penetrating, approaching, and oriented correctly
  colliding_n = jnp.where(penetration < 0., 1., 0.)
  colliding_n *= jnp.where(normal_rel_vel < 0, 1., 0.)
  colliding_n *= jnp.where(impulse > 0., 1., 0.)

  # apply drag if moving laterally above threshold
  colliding_d = colliding_n
  colliding_d *= jnp.where(math.safe_norm(rel_vel_d) > (1. / 100.), 1., 0.)

  return dp_n * colliding_n + dp_d * colliding_d


def _collide_pair(config: config_pb2.Config, body_a: bodies.Body,
                  body_b: bodies.Body, qp_a: QP, qp_b: QP, pos_c: jnp.ndarray,
                  normal_c: jnp.ndarray, penetration: float,
                  dt: float) -> Tuple[P, P]:
  """Calculates velocity change due to a collision.

  Args:
    config: A brax system config.
    body_a: First colliding body
    body_b: Second colliding body
    qp_a: State for first body
    qp_b: State for second body
    pos_c: Where the collision is occuring in world space
    normal_c: Normal vector of surface providing collision
    penetration: Amount of penetration between part and plane
    dt: Integration timestep length

  Returns:
    dP: The impulse on this body result from the collision.
  """
  # TODO: possibly refactor into single collide fn?
  rel_pos_a = pos_c - qp_a.pos
  rel_pos_b = pos_c - qp_b.pos

  vel_a_c = math.world_velocity(qp_a, pos_c)
  vel_b_c = math.world_velocity(qp_b, pos_c)

  rel_vel = vel_b_c - vel_a_c

  baumgarte_rel_vel = config.baumgarte_erp * penetration / dt
  normal_rel_vel = jnp.dot(normal_c, rel_vel)

  temp1 = jnp.matmul(body_a.inertia, jnp.cross(rel_pos_a, normal_c))
  temp2 = jnp.matmul(body_b.inertia, jnp.cross(rel_pos_b, normal_c))

  ang = jnp.dot(normal_c,
                jnp.cross(temp1, rel_pos_a) + jnp.cross(temp2, rel_pos_b))
  impulse = (-1. *
             (1. + config.elasticity) * normal_rel_vel - baumgarte_rel_vel) / (
                 (1. / body_a.mass) + (1. / body_b.mass) + ang)

  dp_n_a = body_a.impulse(qp_a, -impulse * normal_c, pos_c)
  dp_n_b = body_b.impulse(qp_b, impulse * normal_c, pos_c)

  # apply drag due to friction acting parallel to the surface contact
  rel_vel_d = rel_vel - normal_rel_vel * normal_c
  impulse_d = math.safe_norm(rel_vel_d) / ((1. / body_a.mass) +
                                           (1. / body_b.mass) + ang)
  # drag magnitude cannot exceed max friction
  impulse_d = jnp.where(impulse_d < config.friction * impulse, impulse_d,
                        config.friction * impulse)
  dir_d = rel_vel_d / (1e-6 + math.safe_norm(rel_vel_d))
  dp_d_a = body_a.impulse(qp_a, impulse_d * dir_d, pos_c)
  dp_d_b = body_b.impulse(qp_b, -impulse_d * dir_d, pos_c)

  # apply collision normal if penetrating, approaching, and oriented correctly
  colliding_n = jnp.where(penetration < 0., 1., 0.)
  colliding_n *= jnp.where(normal_rel_vel < 0, 1., 0.)
  colliding_n *= jnp.where(impulse > 0., 1., 0.)

  # apply drag if moving laterally above threshold
  colliding_d = colliding_n
  colliding_d *= jnp.where(math.safe_norm(rel_vel_d) > (1. / 100.), 1., 0.)

  dp_a = dp_n_a * colliding_n + dp_d_a * colliding_d
  dp_b = dp_n_b * colliding_n + dp_d_b * colliding_d

  return dp_a, dp_b



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Constraints of joints on body pairs."""

from typing import Any, Tuple

from flax import struct
import jax
import jax.numpy as jnp

from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics import math
from brax.physics.base import P, QP, euler_to_quat, take, vec_to_np

type_to_dof = {"revolute": 1, "universal": 2, "spherical": 3}
lim_to_dof = {0: 1, 1: 1, 2: 2, 3: 3}


@struct.dataclass
class Joint:
  """Connects to bodies together with some axes of rotation."""
  stiffness: jnp.ndarray
  angular_damping: jnp.ndarray
  spring_damping: jnp.ndarray
  limit_strength: jnp.ndarray
  limit: jnp.ndarray
  body_p: bodies.Body
  body_c: bodies.Body
  axis_1: jnp.ndarray
  axis_2: jnp.ndarray
  axis_3: jnp.ndarray
  off_p: jnp.ndarray
  off_c: jnp.ndarray
  ref: jnp.ndarray
  config: config_pb2.Config = struct.field(pytree_node=False)

  @classmethod
  def from_config(cls, config: config_pb2.Config) -> "Joint":
    """Creates a joint from a config."""
    joint_type = cls.__name__.lower()

    joints = [
        j for j in config.joints
        if type_to_dof[joint_type] == lim_to_dof[len(j.angle_limit)]
    ]
    if not joints:
      return cls(*[None] * 14)

    stiffness = jnp.array([j.stiffness for j in joints])
    angular_damping = jnp.array([j.angular_damping for j in joints])

    spring_damping = []
    limit_strength = []
    # these values are empirically determined to be stable defaults
    spring_damping_coeff = {
        "revolute": .5,
        "universal": 2.0,
        "spherical": 2.0
    }[joint_type]

    for j in joints:
      if j.HasField("spring_damping"):
        spring_damping.append(j.spring_damping)
      else:
        spring_damping.append(spring_damping_coeff * jnp.sqrt(j.stiffness))

      if j.HasField("limit_strength"):
        limit_strength.append(j.limit_strength)
      else:
        limit_strength.append(j.stiffness)

    spring_damping = jnp.array(spring_damping)
    limit_strength = jnp.array(limit_strength)

    limits = []
    for j in joints:
      limits.append([[i.min, i.max] for i in j.angle_limit])
    limit = jnp.array(limits) / 180.0 * jnp.pi

    # joint bodies
    body_map = {b.name: i for i, b in enumerate(config.bodies)}
    body = bodies.Body.from_config(config)
    body_p = take(body, jnp.array([body_map[j.parent] for j in joints]))
    body_c = take(body, jnp.array([body_map[j.child] for j in joints]))

    # joint axes
    v_rot = jax.vmap(math.rotate, in_axes=[0, None])
    joint_axes = jnp.round(
        jnp.array(
            [v_rot(jnp.eye(3), euler_to_quat(j.rotation)) for j in joints]),
        decimals=5)
    axis_1 = joint_axes[:, 0]
    axis_2 = joint_axes[:, 1]
    axis_3 = joint_axes[:, 2]
    if joint_type not in ("revolute", "universal", "spherical"):
      raise RuntimeError(f"joint type not implemented { joint_type }")

    # joint offsets
    off_p = jnp.array([vec_to_np(j.parent_offset) for j in joints])
    off_c = jnp.array([vec_to_np(j.child_offset) for j in joints])

    # vector orthogonal to joint axis
    ref_axes = [jnp.array([-a[1], a[0], 0.]) for a in axis_1]
    # one of these two vectors is guaranteed to be orthogonal to axis
    for i, (ax1, ref_axis, ax2) in enumerate(zip(axis_1, ref_axes, axis_2)):
      if joint_type == "universal":
        ref_axes[i] = jnp.cross(ax2, ax1)
      elif jnp.abs(jnp.dot(ax1, ref_axis)) > 1e-4 or jnp.allclose(
          ref_axis, 0, atol=1e-3):
        ref_axes[i] = jnp.array([-ax1[2], 0., ax1[0]])
    ref = jnp.array([r / (1e-6 + jnp.linalg.norm(r)) for r in ref_axes])

    return cls(stiffness, angular_damping, spring_damping, limit_strength,
               limit, body_p, body_c, axis_1, axis_2, axis_3, off_p, off_c, ref,
               config)

  @jax.vmap
  def _apply(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    raise NotImplementedError()  # child must override

  def apply(self, qp: QP) -> P:
    """Constrains and aligns bodies connected by a joint.

    Args:
      qp: State data for system

    Returns:
      dP: Impulses on all bodies to maintain joint constraints
    """
    if not self.config:
      return P(jnp.zeros_like(qp.vel), jnp.zeros_like(qp.ang))

    qp_p = take(qp, self.body_p.idx)
    qp_c = take(qp, self.body_c.idx)
    dp_p, dp_c = self._apply(qp_p, qp_c)

    # sum together all impulse contributions across parents and children
    body_idx = jnp.concatenate((self.body_p.idx, self.body_c.idx))
    dp_vel = jnp.concatenate((dp_p.vel, dp_c.vel))
    dp_ang = jnp.concatenate((dp_p.ang, dp_c.ang))
    dp_vel = jax.ops.segment_sum(dp_vel, body_idx, len(self.config.bodies))
    dp_ang = jax.ops.segment_sum(dp_ang, body_idx, len(self.config.bodies))

    return P(vel=dp_vel, ang=dp_ang)

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint.

    vmap across axis_angle to get all joints.

    Args:
      qp_p: State for parent body
      qp_c: State for child body

    Returns:
      axis: n-tuple of joint axes where n = # DoF of the joint
      angle: n-tuple of joint angles where n = # DoF of the joint
    """
    raise NotImplementedError()  # child must override

  def angle_vel(self, qp: QP) -> Tuple[Any, Any]:
    """Returns joint angle and velocity.

    Args:
      qp: State data for system

    Returns:
      angle: n-tuple of joint angles where n = # DoF of the joint
      vel: n-tuple of joint velocities where n = # DoF of the joint
    """
    if not self.config:
      return None, None

    @jax.vmap
    def op(joint, qp_p, qp_c):
      axes, angles = joint.axis_angle(qp_p, qp_c)
      vels = tuple([jnp.dot(qp_p.ang - qp_c.ang, axis) for axis in axes])
      return angles, vels

    qp_p = take(qp, self.body_p.idx)
    qp_c = take(qp, self.body_c.idx)
    angles, vels = op(self, qp_p, qp_c)

    return angles, vels


@struct.dataclass
class Revolute(Joint):
  """Connects to bodies together with a single axis of rotation."""

  @jax.vmap
  def _apply(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = math.to_world(qp_p, self.off_p)
    pos_c, vel_c = math.to_world(qp_c, self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to align their axes
    (axis,), (angle,) = self.axis_angle(qp_p, qp_c)
    axis_c = math.rotate(self.axis_1, qp_c.rot)
    torque = self.stiffness * jnp.cross(axis, axis_c)

    # torque the bodies to stay within angle limits
    dang = jnp.where(angle < self.limit[0][0], self.limit[0][0] - angle, 0)
    dang = jnp.where(angle > self.limit[0][1], self.limit[0][1] - angle, dang)
    torque -= self.limit_strength * axis * dang

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dang_p = jnp.matmul(self.body_p.inertia, torque)
    dang_c = jnp.matmul(self.body_c.inertia, -torque)
    dp_p = dp_p.replace(ang=dp_p.ang + dang_p)
    dp_c = dp_c.replace(ang=dp_c.ang + dang_c)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    axis_p = math.rotate(self.axis_1, qp_p.rot)
    angle = math.signed_angle(qp_p, qp_c, axis_p, self.ref)
    return (axis_p,), (angle,)


@struct.dataclass
class Universal(Joint):
  """Connects to bodies together with two axes of rotation."""

  @jax.vmap
  def _apply(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = math.to_world(qp_p, self.off_p)
    pos_c, vel_c = math.to_world(qp_c, self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to align to a joint plane
    (axis_1, axis_2), (angle_1, angle_2) = self.axis_angle(qp_p, qp_c)
    axis_c_proj = axis_2 - jnp.dot(axis_2, axis_1) * axis_1
    axis_c_proj = axis_c_proj / math.safe_norm(axis_c_proj)
    torque = (self.limit_strength / 5.) * jnp.cross(axis_c_proj, axis_2)

    # torque the bodies to stay within angle limits
    limit_1, limit_2 = self.limit
    dang_1 = jnp.where(angle_1 < limit_1[0], limit_1[0] - angle_1, 0)
    dang_1 = jnp.where(angle_1 > limit_1[1], limit_1[1] - angle_1, dang_1)
    dang_2 = jnp.where(angle_2 < limit_2[0], limit_2[0] - angle_2, 0)
    dang_2 = jnp.where(angle_2 > limit_2[1], limit_2[1] - angle_2, dang_2)
    torque -= self.limit_strength * (axis_1 * dang_1 + axis_2 * dang_2)

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dang_p = jnp.matmul(self.body_p.inertia, torque)
    dang_c = jnp.matmul(self.body_c.inertia, -torque)
    dp_p = dp_p.replace(ang=dp_p.ang + dang_p)
    dp_c = dp_c.replace(ang=dp_c.ang + dang_c)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    axis_1_c = math.rotate(self.axis_1, qp_c.rot)
    axis_2_p = math.rotate(self.axis_2, qp_p.rot)
    child_in_plane = jnp.cross(axis_2_p, axis_1_c)
    ref_c = math.rotate(self.ref, qp_c.rot)
    angle_1 = math.signed_angle(qp_p, qp_c, axis_2_p, self.ref)
    angle_2 = jnp.arctan2(
        jnp.dot(jnp.cross(child_in_plane, ref_c), axis_1_c),
        jnp.dot(child_in_plane, ref_c))
    return (axis_2_p, axis_1_c), (angle_1, angle_2)


@struct.dataclass
class Spherical(Joint):
  """Connects to bodies together with three axes of rotation."""

  @jax.vmap
  def _apply(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = math.to_world(qp_p, self.off_p)
    pos_c, vel_c = math.to_world(qp_c, self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to stay within angle limits
    axis, angle = self.axis_angle(qp_p, qp_c)
    angle_1, angle_2, angle_3 = angle
    limit_1, limit_2, limit_3 = self.limit
    dang_1 = jnp.where(angle_1 < limit_1[0], limit_1[0] - angle_1, 0)
    dang_1 = jnp.where(angle_1 > limit_1[1], limit_1[1] - angle_1, dang_1)
    dang_2 = jnp.where(angle_2 < limit_2[0], limit_2[0] - angle_2, 0)
    dang_2 = jnp.where(angle_2 > limit_2[1], limit_2[1] - angle_2, dang_2)
    dang_3 = jnp.where(angle_3 < limit_3[0], limit_3[0] - angle_3, 0)
    dang_3 = jnp.where(angle_3 > limit_3[1], limit_3[1] - angle_3, dang_3)

    # TODO: fully decouple different torque axes
    torque = axis[0] * dang_1 + axis[1] * dang_2 + axis[2] * dang_3
    torque *= self.limit_strength

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dp_p = dp_p.replace(ang=dp_p.ang +
                        jnp.matmul(self.body_p.inertia, torque))
    dp_c = dp_c.replace(ang=dp_c.ang +
                        jnp.matmul(self.body_c.inertia, -torque))

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    axis_1_p = math.rotate(self.axis_1, qp_p.rot)
    axis_2_p = math.rotate(self.axis_2, qp_p.rot)
    axis_3_p = math.rotate(self.axis_3, qp_p.rot)
    axis_1_c = math.rotate(self.axis_1, qp_c.rot)
    axis_2_c = math.rotate(self.axis_2, qp_c.rot)
    axis_3_c = math.rotate(self.axis_3, qp_c.rot)

    axis_2_in_plane = axis_2_c - jnp.dot(axis_2_c, axis_3_p) * axis_3_p
    axis_2_in_projected_length = jnp.linalg.norm(axis_2_in_plane)
    axis_2_in_plane = axis_2_in_plane / (1e-7 +
                                         jnp.linalg.norm(axis_2_in_plane))

    angle_z = jnp.arctan2(
        jnp.dot(axis_2_in_plane, axis_1_p),
        jnp.dot(axis_2_in_plane, axis_2_p))  # parent z axis torque
    angle_x = -1. * jnp.arctan2(
        jnp.dot(axis_2_c, axis_3_p),
        axis_2_in_projected_length)  # child x axis torque
    axis_3_in_child_xz = axis_3_p - jnp.dot(axis_3_p, axis_2_c) * axis_2_c
    axis_3_in_child_xz = axis_3_in_child_xz / (
        1e-7 + jnp.linalg.norm(axis_3_in_child_xz))
    angle_y = jnp.arctan2(
        jnp.dot(
            axis_3_in_child_xz -
            jnp.dot(axis_3_in_child_xz, axis_3_c) * axis_3_c, axis_1_c),
        jnp.dot(
            axis_3_in_child_xz -
            jnp.dot(axis_3_in_child_xz, axis_1_c) * axis_1_c,
            axis_3_c))  # child y torque

    axis = (axis_1_c, axis_2_c, axis_3_p,)
    angle = (angle_x, angle_y, angle_z)

    return axis, angle



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.envs."""

import logging
import time

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
import jax
import jax.numpy as np

_EARLY_TERMINATION = ('ant', 'humanoid')
_EXPECTED_SPS = {'ant': 1000, 'fetch': 1000}


class EnvTest(parameterized.TestCase):

  @parameterized.parameters(_EXPECTED_SPS.items())
  def testSpeed(self, env_name, expected_sps):
    batch_size = 128
    episode_length = 1000
    auto_reset = (env_name not in _EARLY_TERMINATION)

    env = envs.create(
        env_name,
        episode_length=episode_length,
        auto_reset=auto_reset,
        batch_size=batch_size)
    zero_action = np.zeros((batch_size, env.action_size))

    @jax.jit
    def run_env(state):

      def step(carry, _):
        state, = carry
        state = env.step(state, zero_action)
        return (state,), ()

      (state,), _ = jax.lax.scan(step, (state,), (), length=episode_length)
      return state

    # warmup
    rng = jax.random.PRNGKey(0)
    state = env.reset(rng)
    state = run_env(state)
    state.done.block_until_ready()

    sps = []
    for seed in range(5):
      rng = jax.random.PRNGKey(seed)
      state = env.reset(rng)
      jax.device_put(state)
      t = time.time()
      state = run_env(state)
      state.done.block_until_ready()
      sps.append((batch_size * episode_length) / (time.time() - t))
      self.assertTrue(np.alltrue(state.done))

    mean_sps = np.mean(np.array(sps))
    logging.info('%s SPS %s %s', env_name, mean_sps, sps)
    self.assertGreater(mean_sps, expected_sps * 0.99)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ARS training tests."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax.training import ars
from flax import serialization
import jax


class ARSTest(parameterized.TestCase):
  """Tests for ARS module."""


  @parameterized.parameters(True, False)
  def testModelEncoding(self, normalize_observations):
    env_fn = envs.create_fn('fast')
    _, params, _ = ars.train(
        env_fn,
        num_timesteps=128,
        episode_length=128)
    env = env_fn()
    base_params, inference = ars.make_params_and_inference_fn(
        env.observation_size, env.action_size, normalize_observations)
    byte_encoding = serialization.to_bytes(params)
    decoded_params = serialization.from_bytes(base_params, byte_encoding)

    # Compute one action.
    state = env.reset(jax.random.PRNGKey(0))
    action = inference(decoded_params, state.obs, jax.random.PRNGKey(0))
    env.step(state, action)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Saves a system config and trajectory as json."""

import json
from typing import List

import jax.numpy as np
import brax
from brax.io.file import File

from google.protobuf import json_format


class JaxEncoder(json.JSONEncoder):

  def default(self, obj):
    if isinstance(obj, np.ndarray):
      return obj.tolist()
    return json.JSONEncoder.default(self, obj)


def save(path: str, sys: brax.System, qps: List[brax.QP]):
  with File(path, 'w') as fout:
    d = {'config': json_format.MessageToDict(sys.config, True),
         'pos': [qp.pos for qp in qps],
         'rot': [qp.rot for qp in qps],}
    fout.write(json.dumps(d, cls=JaxEncoder))



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proximal policy optimization training.

See: https://arxiv.org/pdf/1707.06347.pdf
"""

import functools
import time
from typing import Any, Callable, Dict, Optional, Tuple

from absl import logging
from brax import envs
from brax.training import distribution
from brax.training import env
from brax.training import networks
from brax.training import normalization
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import optax


def compute_gae(truncation: jnp.ndarray,
                rewards: jnp.ndarray,
                values: jnp.ndarray,
                bootstrap_value: jnp.ndarray,
                lambda_: float = 1.0,
                discount: float = 0.99):
  r"""Calculates the Generalized Advantage Estimation (GAE).

  Args:
    truncation: A float32 tensor of shape [T, B] with truncation signal.
    rewards: A float32 tensor of shape [T, B] containing rewards generated by
      following the behaviour policy.
    values: A float32 tensor of shape [T, B] with the value function estimates
      wrt. the target policy.
    bootstrap_value: A float32 of shape [B] with the value function estimate at
      time T.
    lambda_: Mix between 1-step (lambda_=0) and n-step (lambda_=1). Defaults to
      lambda_=1.
    discount: TD discount.

  Returns:
    A float32 tensor of shape [T, B]. Can be used as target to
      train a baseline (V(x_t) - vs_t)^2.
    A float32 tensor of shape [T, B] of advantages.
  """

  truncation_mask = 1 - truncation
  # Append bootstrapped value to get [v1, ..., v_t+1]
  values_t_plus_1 = jnp.concatenate(
      [values[1:], jnp.expand_dims(bootstrap_value, 0)], axis=0)
  deltas = rewards + discount * values_t_plus_1 - values
  deltas *= truncation_mask

  acc = jnp.zeros_like(bootstrap_value)
  vs_minus_v_xs = []

  def compute_vs_minus_v_xs(carry, target_t):
    lambda_, acc = carry
    truncation_mask, delta = target_t
    acc = delta + discount * truncation_mask * lambda_ * acc
    return (lambda_, acc), (acc)

  (_, _), (vs_minus_v_xs) = jax.lax.scan(compute_vs_minus_v_xs, (lambda_, acc),
                                         (truncation_mask, deltas),
                                         length=int(truncation_mask.shape[0]),
                                         reverse=True)
  # Add V(x_s) to get v_s.
  vs = jnp.add(vs_minus_v_xs, values)

  vs_t_plus_1 = jnp.concatenate([vs[1:], jnp.expand_dims(bootstrap_value, 0)],
                                axis=0)
  advantages = (rewards + discount * vs_t_plus_1 - values) * truncation_mask
  return jax.lax.stop_gradient(vs), jax.lax.stop_gradient(advantages)


@flax.struct.dataclass
class StepData:
  """Contains data for one environment step."""
  obs: jnp.ndarray
  rewards: jnp.ndarray
  dones: jnp.ndarray
  actions: jnp.ndarray
  logits: jnp.ndarray


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  optimizer_state: optax.OptState
  params: Params
  key: PRNGKey
  normalizer_params: Params


def compute_ppo_loss(
    models: Dict[str, Params],
    data: StepData,
    rng: jnp.ndarray,
    parametric_action_distribution: distribution.ParametricDistribution,
    policy_apply: Any,
    value_apply: Any,
    entropy_cost: float = 1e-4,
    discounting: float = 0.9,
    reward_scaling: float = 1.0,
    lambda_: float = 0.95,
    ppo_epsilon: float = 0.3):
  """Computes PPO loss."""
  policy_params, value_params = models['policy'], models['value']
  policy_logits = policy_apply(policy_params, data.obs[:-1])
  baseline = value_apply(value_params, data.obs)
  baseline = jnp.squeeze(baseline, axis=-1)

  # Use last baseline value (from the value function) to bootstrap.
  bootstrap_value = baseline[-1]
  baseline = baseline[:-1]

  # At this point, we have unroll length + 1 steps. The last step is only used
  # as bootstrap value, so it's removed.

  # already removed at data generation time
  # actions = actions[:-1]
  # logits = logits[:-1]

  rewards = data.rewards[1:] * reward_scaling
  truncation = data.dones[1:]

  target_action_log_probs = parametric_action_distribution.log_prob(
      policy_logits, data.actions)
  behaviour_action_log_probs = parametric_action_distribution.log_prob(
      data.logits, data.actions)

  vs, advantages = compute_gae(
      truncation=truncation,
      rewards=rewards,
      values=baseline,
      bootstrap_value=bootstrap_value,
      lambda_=lambda_,
      discount=discounting)
  rho_s = jnp.exp(target_action_log_probs - behaviour_action_log_probs)

  surrogate_loss1 = rho_s * advantages
  surrogate_loss2 = jnp.clip(rho_s, 1 - ppo_epsilon,
                             1 + ppo_epsilon) * advantages

  policy_loss = -jnp.mean(jnp.minimum(surrogate_loss1, surrogate_loss2))

  # Value function loss
  v_error = vs - baseline
  v_loss = jnp.mean(v_error * v_error) * 0.5 * 0.5

  # Entropy reward
  entropy = jnp.mean(
      parametric_action_distribution.entropy(policy_logits, rng))
  entropy_loss = entropy_cost * -entropy

  return policy_loss + v_loss + entropy_loss, {
      'total_loss': policy_loss + v_loss + entropy_loss,
      'policy_loss': policy_loss,
      'v_loss': v_loss,
      'entropy_loss': entropy_loss
  }


def train(
    environment_fn: Callable[..., envs.Env],
    num_timesteps,
    episode_length: int,
    action_repeat: int = 1,
    num_envs: int = 1,
    max_devices_per_host: Optional[int] = None,
    num_eval_envs: int = 128,
    learning_rate=1e-4,
    entropy_cost=1e-4,
    discounting=0.9,
    seed=0,
    unroll_length=10,
    batch_size=32,
    num_minibatches=16,
    num_update_epochs=2,
    log_frequency=10,
    normalize_observations=False,
    reward_scaling=1.,
    progress_fn: Optional[Callable[[int, Dict[str, Any]], None]] = None,
):
  """PPO training."""
  assert batch_size * num_minibatches % num_envs == 0
  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d',
      jax.device_count(), process_count, process_id, local_device_count,
      local_devices_to_use)

  key = jax.random.PRNGKey(seed)
  key, key_models, key_env, key_eval = jax.random.split(key, 4)
  # Make sure every process gets a different random key, otherwise they will be
  # doing identical work.
  key_env = jax.random.split(key_env, process_count)[process_id]
  key = jax.random.split(key, process_count)[process_id]
  # key_models should be the same, so that models are initialized the same way
  # for different processes

  core_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_envs // local_devices_to_use // process_count,
      episode_length=episode_length)
  key_envs = jax.random.split(key_env, local_devices_to_use)
  step_fn = core_env.step
  first_state = jax.tree_multimap(lambda *args: jnp.stack(args),
                                  *[core_env.reset(key) for key in key_envs])

  core_eval_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_eval_envs,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = env.wrap_for_eval(core_eval_env, key_eval)

  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=core_env.action_size)

  policy_model, value_model = networks.make_models(
      parametric_action_distribution.param_size,
      core_env.observation_size)
  key_policy, key_value = jax.random.split(key_models)

  optimizer = optax.adam(learning_rate=learning_rate)
  init_params = {'policy': policy_model.init(key_policy),
                 'value': value_model.init(key_value)}
  optimizer_state = optimizer.init(init_params)
  optimizer_state, init_params = normalization.bcast_local_devices(
      (optimizer_state, init_params), local_devices_to_use)

  normalizer_params, obs_normalizer_update_fn, obs_normalizer_apply_fn = (
      normalization.create_observation_normalizer(
          core_env.observation_size, normalize_observations,
          num_leading_batch_dims=2, pmap_to_devices=local_devices_to_use))

  key_debug = jax.random.PRNGKey(seed + 666)

  loss_fn = functools.partial(
      compute_ppo_loss,
      parametric_action_distribution=parametric_action_distribution,
      policy_apply=policy_model.apply,
      value_apply=value_model.apply,
      entropy_cost=entropy_cost,
      discounting=discounting,
      reward_scaling=reward_scaling)

  grad_loss = jax.grad(loss_fn, has_aux=True)

  def do_one_step_eval(carry, unused_target_t):
    state, policy_params, normalizer_params, key = carry
    key, key_sample = jax.random.split(key)
    # TODO: Make this nicer ([0] comes from pmapping).
    obs = obs_normalizer_apply_fn(
        jax.tree_map(lambda x: x[0], normalizer_params), state.core.obs)
    logits = policy_model.apply(policy_params, obs)
    actions = parametric_action_distribution.sample(logits, key_sample)
    nstate = eval_step_fn(state, actions)
    return (nstate, policy_params, normalizer_params, key), ()

  @jax.jit
  def run_eval(state, key, policy_params,
               normalizer_params) -> Tuple[env.EvalEnvState, PRNGKey]:
    policy_params = jax.tree_map(lambda x: x[0], policy_params)
    (state, _, _, key), _ = jax.lax.scan(
        do_one_step_eval, (state, policy_params, normalizer_params, key), (),
        length=episode_length // action_repeat)
    return state, key

  def do_one_step(carry, unused_target_t):
    state, normalizer_params, policy_params, key = carry
    key, key_sample = jax.random.split(key)
    normalized_obs = obs_normalizer_apply_fn(normalizer_params, state.obs)
    logits = policy_model.apply(policy_params, normalized_obs)
    actions = parametric_action_distribution.sample_no_postprocessing(
        logits, key_sample)
    postprocessed_actions = parametric_action_distribution.postprocess(actions)
    nstate = step_fn(state, postprocessed_actions)
    return (nstate, normalizer_params, policy_params, key), StepData(
        obs=state.obs,
        rewards=state.reward,
        dones=state.done,
        actions=actions,
        logits=logits)

  def generate_unroll(carry, unused_target_t):
    state, normalizer_params, policy_params, key = carry
    (state, _, _, key), data = jax.lax.scan(
        do_one_step, (state, normalizer_params, policy_params, key), (),
        length=unroll_length)
    data = data.replace(
        obs=jnp.concatenate(
            [data.obs, jnp.expand_dims(state.obs, axis=0)]),
        rewards=jnp.concatenate(
            [data.rewards, jnp.expand_dims(state.reward, axis=0)]),
        dones=jnp.concatenate(
            [data.dones, jnp.expand_dims(state.done, axis=0)]))
    return (state, normalizer_params, policy_params, key), data

  def update_model(carry, data):
    optimizer_state, params, key = carry
    key, key_loss = jax.random.split(key)
    loss_grad, metrics = grad_loss(params, data, key_loss)
    loss_grad = jax.lax.pmean(loss_grad, axis_name='i')

    params_update, optimizer_state = optimizer.update(loss_grad,
                                                      optimizer_state)
    params = optax.apply_updates(params, params_update)

    return (optimizer_state, params, key), metrics

  def minimize_epoch(carry, unused_t):
    optimizer_state, params, data, key = carry
    key, key_perm, key_grad = jax.random.split(key, 3)
    permutation = jax.random.permutation(key_perm, data.obs.shape[1])
    def convert_data(data, permutation):
      data = jnp.take(data, permutation, axis=1, mode='clip')
      data = jnp.reshape(
          data, [data.shape[0], num_minibatches, -1] + list(data.shape[2:]))
      data = jnp.swapaxes(data, 0, 1)
      return data
    ndata = jax.tree_map(lambda x: convert_data(x, permutation), data)
    (optimizer_state, params, _), metrics = jax.lax.scan(
        update_model, (optimizer_state, params, key_grad),
        ndata,
        length=num_minibatches)
    return (optimizer_state, params, data, key), metrics

  def run_epoch(carry: Tuple[TrainingState, envs.State], unused_t):
    training_state, state = carry
    key_minimize, key_generate_unroll, new_key = jax.random.split(
        training_state.key, 3)
    (state, _, _, _), data = jax.lax.scan(
        generate_unroll, (state, training_state.normalizer_params,
                          training_state.params['policy'],
                          key_generate_unroll), (),
        length=batch_size * num_minibatches // num_envs)
    # make unroll first
    data = jax.tree_map(lambda x: jnp.swapaxes(x, 0, 1), data)
    data = jax.tree_map(
        lambda x: jnp.reshape(x, [x.shape[0], -1] + list(x.shape[3:])), data)

    # Update normalization params and normalize observations.
    normalizer_params = obs_normalizer_update_fn(
        training_state.normalizer_params, data.obs[:-1])
    data = data.replace(
        obs=obs_normalizer_apply_fn(normalizer_params, data.obs))

    (optimizer_state, params, _, _), metrics = jax.lax.scan(
        minimize_epoch, (training_state.optimizer_state, training_state.params,
                         data, key_minimize), (),
        length=num_update_epochs)

    new_training_state = TrainingState(
        optimizer_state=optimizer_state,
        params=params,
        normalizer_params=normalizer_params,
        key=new_key)
    return (new_training_state, state), metrics

  num_epochs = num_timesteps // (
      batch_size * unroll_length * num_minibatches * action_repeat)

  def _minimize_loop(training_state, state):
    (training_state, state), losses = jax.lax.scan(
        run_epoch, (training_state, state), (),
        length=num_epochs // log_frequency)
    losses = jax.tree_map(jnp.mean, losses)
    return (training_state, state), losses

  minimize_loop = jax.pmap(_minimize_loop, axis_name='i')

  training_state = TrainingState(
      optimizer_state=optimizer_state,
      params=init_params,
      key=jnp.stack(jax.random.split(key, local_devices_to_use)),
      normalizer_params=normalizer_params)
  training_walltime = 0
  eval_walltime = 0
  sps = 0
  eval_sps = 0
  losses = {}
  state = first_state
  metrics = {}

  for it in range(log_frequency + 1):
    logging.info('starting iteration %s %s', it, time.time() - xt)
    t = time.time()

    if process_id == 0:
      eval_state, key_debug = (
          run_eval(eval_first_state, key_debug,
                   training_state.params['policy'],
                   training_state.normalizer_params))
      eval_state.completed_episodes.block_until_ready()
      eval_walltime += time.time() - t
      eval_sps = (episode_length * eval_first_state.core.reward.shape[0] /
                  (time.time() - t))
      avg_episode_length = (
          eval_state.completed_episodes_steps / eval_state.completed_episodes)
      metrics = dict(
          dict({
              f'eval/episode_{name}': value / eval_state.completed_episodes
              for name, value in eval_state.completed_episodes_metrics.items()
          }),
          **dict({
              f'losses/{name}': jnp.mean(value)
              for name, value in losses.items()
          }),
          **dict({
              'eval/completed_episodes': eval_state.completed_episodes,
              'eval/avg_episode_length': avg_episode_length,
              'speed/sps': sps,
              'speed/eval_sps': eval_sps,
              'speed/training_walltime': training_walltime,
              'speed/eval_walltime': eval_walltime,
              'speed/timestamp': training_walltime,
          }))
      logging.info(metrics)
      if progress_fn:
        progress_fn(int(training_state.normalizer_params[0][0]) * action_repeat,
                    metrics)

    if it == log_frequency:
      break

    t = time.time()
    previous_step = training_state.normalizer_params[0][0]
    # optimization
    (training_state, state), losses = minimize_loop(training_state, state)
    jax.tree_map(lambda x: x.block_until_ready(), losses)
    sps = ((training_state.normalizer_params[0][0] - previous_step) /
           (time.time() - t)) * action_repeat
    training_walltime += time.time() - t

  # To undo the pmap.
  normalizer_params = jax.tree_map(lambda x: x[0],
                                   training_state.normalizer_params)
  policy_params = jax.tree_map(lambda x: x[0],
                               training_state.params['policy'])

  logging.info('total steps: %s', normalizer_params[0] * action_repeat)

  _, inference = make_params_and_inference_fn(core_env.observation_size,
                                              core_env.action_size,
                                              normalize_observations)
  params = normalizer_params, policy_params

  if process_count > 1:
    # Make sure all processes stay up until the end of main.
    x = jnp.ones([jax.local_device_count()])
    x = jax.device_get(jax.pmap(lambda x: jax.lax.psum(x, 'i'), 'i')(x))
    assert x[0] == jax.device_count()

  return (inference, params, metrics)


def make_params_and_inference_fn(observation_size, action_size,
                                 normalize_observations):
  """Creates params and inference function for the PPO agent."""
  obs_normalizer_params, obs_normalizer_apply_fn = normalization.make_data_and_apply_fn(
      observation_size, normalize_observations)
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size)
  policy_model, _ = networks.make_models(
      parametric_action_distribution.param_size,
      observation_size)

  def inference_fn(params, obs, key):
    normalizer_params, policy_params = params
    obs = obs_normalizer_apply_fn(normalizer_params, obs)
    action = parametric_action_distribution.sample(
        policy_model.apply(policy_params, obs), key)
    return action

  params = (obs_normalizer_params, policy_model.init(jax.random.PRNGKey(0)))
  return params, inference_fn



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains an ant to run in the +x direction."""

import brax
from brax.envs import env
import jax.numpy as jnp


class Ant(env.Env):
  """Trains an ant to run in the +x direction."""

  def __init__(self, **kwargs):
    super().__init__(_SYSTEM_CONFIG, **kwargs)

  def reset(self, rng: jnp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    qp = self.sys.default_qp()
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done = jnp.zeros(2)
    return env.State(qp, obs, reward, done)

  def step(self, state: env.State, action: jnp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    x_before = state.qp.pos[0, 0]
    x_after = qp.pos[0, 0]
    forward_reward = (x_after - x_before) / self.sys.config.dt
    ctrl_cost = .5 * jnp.sum(jnp.square(action))
    # can ignore contact cost, see: https://github.com/openai/gym/issues/1541
    contact_cost = 0
    survive_reward = 1.0
    reward = forward_reward - ctrl_cost - contact_cost + survive_reward

    done = jnp.where(qp.pos[0, 2] < 0.2, x=1.0, y=0.0)
    done = jnp.where(qp.pos[0, 2] > 1.0, x=1.0, y=done)

    return state.replace(qp=qp, obs=obs, reward=reward, done=done)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jnp.ndarray:
    """Observe ant body position and velocities."""
    # some pre-processing to pull joint angles and velocities
    (joint_angle,), (joint_vel,) = self.sys.joint_revolute.angle_vel(qp)

    # qpos:
    # Z of the torso (1,)
    # orientation of the torso as quaternion (4,)
    # joint angles (8,)
    qpos = [qp.pos[0, 2:], qp.rot[0], joint_angle]

    # qvel:
    # velcotiy of the torso (3,)
    # angular velocity of the torso (3,)
    # joint angle velocities (8,)
    qvel = [qp.vel[0], qp.ang[0], joint_vel]

    # external contact forces:
    # delta velocity (3,), delta ang (3,) * 10 bodies in the system
    # Note that mujoco has 4 extra bodies tucked inside the Torso that Brax
    # ignores
    cfrc = [
        jnp.clip(info.contact.vel, -1, 1),
        jnp.clip(info.contact.ang, -1, 1)
    ]
    # flatten bottom dimension
    cfrc = [jnp.reshape(x, x.shape[:-2] + (-1,)) for x in cfrc]

    return jnp.concatenate(qpos + qvel + cfrc)

_SYSTEM_CONFIG = """
bodies {
  name: "$ Torso"
  colliders {
    capsule {
      radius: 0.25
      length: 0.5
      end: 1
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 10
}
bodies {
  name: "Aux 1"
  colliders {
    rotation { x: 90 y: -45 }
    capsule {
      radius: 0.08
      length: 0.4428427219390869
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "$ Body 4"
  colliders {
    rotation { x: 90 y: -45 }
    capsule {
      radius: 0.08
      length: 0.7256854176521301
      end: -1
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "Aux 2"
  colliders {
    rotation { x: 90 y: 45 }
    capsule {
      radius: 0.08
      length: 0.4428427219390869
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "$ Body 7"
  colliders {
    rotation { x: 90 y: 45 }
    capsule {
      radius: 0.08
      length: 0.7256854176521301
      end: -1
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "Aux 3"
  colliders {
    rotation { x: -90 y: 45 }
    capsule {
      radius: 0.08
      length: 0.4428427219390869
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "$ Body 10"
  colliders {
    rotation { x: -90 y: 45 }
    capsule {
      radius: 0.08
      length: 0.7256854176521301
      end: -1
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "Aux 4"
  colliders {
    rotation { x: -90 y: -45 }
    capsule {
      radius: 0.08
      length: 0.4428427219390869
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "$ Body 13"
  colliders {
    rotation { x: -90 y: -45 }
    capsule {
      radius: 0.08
      length: 0.7256854176521301
      end: -1
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
}
bodies {
  name: "Ground"
  colliders {
    plane {}
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 1
  frozen { all: true }
}
joints {
  name: "$ Torso_Aux 1"
  parent_offset { x: 0.2 y: 0.2 }
  child_offset { x: -0.1 y: -0.1 }
  parent: "$ Torso"
  child: "Aux 1"
  stiffness: 5000.0
  angular_damping: 35
  angle_limit { min: -30.0 max: 30.0 }
  rotation { y: -90 }
}
joints {
  name: "Aux 1_$ Body 4"
  parent_offset { x: 0.1 y: 0.1 }
  child_offset { x: -0.2 y: -0.2 }
  parent: "Aux 1"
  child: "$ Body 4"
  stiffness: 5000.0
  angular_damping: 35
  rotation: { z: 135 }
  angle_limit {
    min: 30.0
    max: 70.0
  }
}
joints {
  name: "$ Torso_Aux 2"
  parent_offset { x: -0.2 y: 0.2 }
  child_offset { x: 0.1 y: -0.1 }
  parent: "$ Torso"
  child: "Aux 2"
  stiffness: 5000.0
  angular_damping: 35
  rotation { y: -90 }
  angle_limit { min: -30.0 max: 30.0 }
}
joints {
  name: "Aux 2_$ Body 7"
  parent_offset { x: -0.1 y: 0.1 }
  child_offset { x: 0.2 y: -0.2 }
  parent: "Aux 2"
  child: "$ Body 7"
  stiffness: 5000.0
  angular_damping: 35
  rotation { z: 45 }
  angle_limit { min: -70.0 max: -30.0 }
}
joints {
  name: "$ Torso_Aux 3"
  parent_offset { x: -0.2 y: -0.2 }
  child_offset { x: 0.1 y: 0.1 }
  parent: "$ Torso"
  child: "Aux 3"
  stiffness: 5000.0
  angular_damping: 35
  rotation { y: -90 }
  angle_limit { min: -30.0 max: 30.0 }
}
joints {
  name: "Aux 3_$ Body 10"
  parent_offset { x: -0.1 y: -0.1 }
  child_offset {
    x: 0.2
    y: 0.2
  }
  parent: "Aux 3"
  child: "$ Body 10"
  stiffness: 5000.0
  angular_damping: 35
  rotation { z: 135 }
  angle_limit { min: -70.0 max: -30.0 }
}
joints {
  name: "$ Torso_Aux 4"
  parent_offset { x: 0.2 y: -0.2 }
  child_offset { x: -0.1 y: 0.1 }
  parent: "$ Torso"
  child: "Aux 4"
  stiffness: 5000.0
  angular_damping: 35
  rotation { y: -90 }
  angle_limit { min: -30.0 max: 30.0 }
}
joints {
  name: "Aux 4_$ Body 13"
  parent_offset { x: 0.1 y: -0.1 }
  child_offset { x: -0.2 y: 0.2 }
  parent: "Aux 4"
  child: "$ Body 13"
  stiffness: 5000.0
  angular_damping: 35
  rotation { z: 45 }
  angle_limit { min: 30.0 max: 70.0 }
}
actuators {
  name: "$ Torso_Aux 1"
  joint: "$ Torso_Aux 1"
  strength: 300.0
  torque {}
}
actuators {
  name: "Aux 1_$ Body 4"
  joint: "Aux 1_$ Body 4"
  strength: 300.0
  torque {}
}
actuators {
  name: "$ Torso_Aux 2"
  joint: "$ Torso_Aux 2"
  strength: 300.0
  torque {}
}
actuators {
  name: "Aux 2_$ Body 7"
  joint: "Aux 2_$ Body 7"
  strength: 300.0
  torque {}
}
actuators {
  name: "$ Torso_Aux 3"
  joint: "$ Torso_Aux 3"
  strength: 300.0
  torque {}
}
actuators {
  name: "Aux 3_$ Body 10"
  joint: "Aux 3_$ Body 10"
  strength: 300.0
  torque {}
}
actuators {
  name: "$ Torso_Aux 4"
  joint: "$ Torso_Aux 4"
  strength: 300.0
  torque {}
}
actuators {
  name: "Aux 4_$ Body 13"
  joint: "Aux 4_$ Body 13"
  strength: 300.0
  torque {}
}
friction: 1.0
gravity { z: -9.8 }
angular_damping: -0.05
baumgarte_erp: 0.1
collide_include {
  first: "$ Torso"
  second: "Ground"
}
collide_include {
  first: "$ Body 4"
  second: "Ground"
}
collide_include {
  first: "$ Body 7"
  second: "Ground"
}
collide_include {
  first: "$ Body 10"
  second: "Ground"
}
collide_include {
  first: "$ Body 13"
  second: "Ground"
}
dt: 0.05
substeps: 10
"""



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evolution strategy training.

See: https://arxiv.org/pdf/1703.03864.pdf
"""

import time
from typing import Any, Callable, Dict, Optional, Tuple

from absl import logging
from brax import envs
from brax.training import distribution
from brax.training import env
from brax.training import networks
from brax.training import normalization
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import optax


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  key: PRNGKey
  normalizer_params: Params
  optimizer_state: optax.OptState
  policy_params: Params


def train(
    environment_fn: Callable[..., envs.Env],
    num_timesteps: int = 100,
    episode_length: int = 1000,
    action_repeat: int = 1,
    l2coeff: float = 0,
    max_devices_per_host: Optional[int] = None,
    population_size: int = 128,
    learning_rate: float = 0.001,
    fitness_shaping: int = 0,
    num_eval_envs: int = 128,
    perturbation_std: float = 0.1,
    seed: int = 0,
    normalize_observations: bool = False,
    log_frequency: int = 1,
    center_fitness: bool = False,
    progress_fn: Optional[Callable[[int, Dict[str, Any]], None]] = None,
):
  """ES training (from https://arxiv.org/pdf/1703.03864.pdf)."""
  num_envs = population_size * 2  # antitethic

  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d',
      jax.device_count(), process_count, process_id, local_device_count,
      local_devices_to_use)

  key = jax.random.PRNGKey(seed)
  key, key_model, key_env, key_eval = jax.random.split(key, 4)

  core_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_envs // local_devices_to_use // process_count,
      episode_length=episode_length)
  key_envs = jax.random.split(key_env, local_devices_to_use // 2)
  step_fn = core_env.step
  first_state = jax.tree_multimap(lambda *args: jnp.stack(args),
                                  *[core_env.reset(key) for key in key_envs])

  core_eval_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_eval_envs,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = env.wrap_for_eval(core_eval_env, key_eval)

  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=core_env.action_size)

  _, obs_size = eval_first_state.core.obs.shape

  policy_model = make_es_model(parametric_action_distribution, obs_size)

  normalizer_params, obs_normalizer_update_fn, obs_normalizer_apply_fn = (
      normalization.create_observation_normalizer(
          obs_size, normalize_observations, num_leading_batch_dims=2))

  optimizer = optax.adam(learning_rate=learning_rate)
  policy_params = policy_model.init(key_model)
  optimizer_state = optimizer.init(policy_params)

  key_debug = jax.random.PRNGKey(seed + 666)

  def do_one_step_eval(carry, unused_target_t):
    state, policy_params, normalizer_params, key = carry
    key, key_sample = jax.random.split(key)
    obs = obs_normalizer_apply_fn(normalizer_params, state.core.obs)
    logits = policy_model.apply(policy_params, obs)
    actions = parametric_action_distribution.sample(logits, key_sample)
    nstate = eval_step_fn(state, actions)
    return (nstate, policy_params, normalizer_params, key), ()

  @jax.jit
  def run_eval(state, key, policy_params,
               normalizer_params) -> Tuple[env.EvalEnvState, PRNGKey]:
    (state, _, _, key), _ = jax.lax.scan(
        do_one_step_eval, (state, policy_params, normalizer_params, key), (),
        length=episode_length // action_repeat)
    return state, key

  @jax.vmap
  def training_inference(params, obs):
    return policy_model.apply(params, obs)

  def do_one_step(carry, unused_target_t):
    state, policy_params, key, cumulative_reward, active_episode, normalizer_params = carry
    key, key_sample = jax.random.split(key)
    obs = obs_normalizer_apply_fn(normalizer_params, state.obs)
    logits = training_inference(policy_params, obs)
    actions = parametric_action_distribution.sample(logits, key_sample)
    nstate = step_fn(state, actions)
    cumulative_reward = cumulative_reward + nstate.reward * active_episode
    new_active_episode = active_episode * (1 - nstate.done)
    return (nstate, policy_params, key, cumulative_reward, new_active_episode,
            normalizer_params), (state.obs, active_episode)

  def run_es_eval(state, params, key, normalizer_params):
    cumulative_reward = jnp.zeros(state.obs.shape[0])
    active_episode = jnp.ones_like(cumulative_reward)
    (state, _, key, cumulative_reward, _, _), (obs, obs_weights) = jax.lax.scan(
        do_one_step, (state, params, key, cumulative_reward, active_episode,
                      normalizer_params), (),
        length=episode_length // action_repeat)
    return cumulative_reward, obs, obs_weights, state

  def add_noise(params, key):
    num_vars = len(jax.tree_leaves(params))
    treedef = jax.tree_structure(params)
    all_keys = jax.random.split(key, num=num_vars)
    noise = jax.tree_multimap(
        lambda g, k: jax.random.normal(k, shape=g.shape, dtype=g.dtype), params,
        jax.tree_unflatten(treedef, all_keys))
    params_with_noise = jax.tree_multimap(lambda g, n: g + n * perturbation_std,
                                          params, noise)
    anit_params_with_noise = jax.tree_multimap(
        lambda g, n: g - n * perturbation_std, params, noise)
    return params_with_noise, anit_params_with_noise, noise

  @jax.jit
  def es_one_epoch(state: envs.State, training_state: TrainingState):
    params = jax.tree_map(
        lambda x: jnp.repeat(jnp.expand_dims(x, axis=0),
                             population_size, axis=0),
        training_state.policy_params)
    key, key_petr, key_es_eval = jax.random.split(training_state.key, 3)
    # generate perturbations
    params_with_noise, params_with_anti_noise, noise = add_noise(
        params, key_petr)

    pstate = jax.tree_map(lambda x: jnp.concatenate([x, x], axis=0), state)
    pparams = jax.tree_multimap(lambda a, b: jnp.concatenate([a, b], axis=0),
                                params_with_noise, params_with_anti_noise)

    pparams = jax.tree_map(
        lambda x: jnp.reshape(x, [local_devices_to_use, -1] + list(x.shape[1:])
                             ), pparams)

    prun_es_eval = jax.pmap(run_es_eval, in_axes=(0, 0, None, None))
    eval_scores, obs, obs_weights, state = prun_es_eval(
        pstate, pparams, key_es_eval,
        training_state.normalizer_params)

    state = jax.tree_map(lambda x: jnp.split(x, 2, axis=0)[0], state)

    obs = jnp.reshape(obs, [-1] + list(obs.shape[2:]))
    obs_weights = jnp.reshape(obs_weights, [-1] + list(obs_weights.shape[2:]))

    normalizer_params = obs_normalizer_update_fn(
        training_state.normalizer_params, obs, obs_weights)

    weights = jnp.reshape(eval_scores, [-1])

    # aggregate results
    if fitness_shaping == 1:
      # Shaping from
      # https://www.jmlr.org/papers/volume15/wierstra14a/wierstra14a.pdf
      weights = weights.shape[0] - jnp.argsort(jnp.argsort(weights))
      weights = jnp.maximum(
          0,
          jnp.log(weights.shape[0] / 2.0 + 1) - jnp.log(weights))
      weights = weights / jnp.sum(weights) - 1.0 / weights.shape[0]
    elif fitness_shaping == 2:
      # Centered rank from: https://arxiv.org/pdf/1703.03864.pdf
      weights = jnp.argsort(jnp.argsort(weights))
      weights /= (weights.shape[0] - 1)
      weights -= .5
    elif fitness_shaping == 0:
      # Original scores.
      pass
    else:
      assert 0

    if center_fitness:
      weights = (weights - jnp.mean(weights)) / (1e-6 + jnp.std(weights))

    weights1, weights2 = jnp.split(weights, 2)
    weights = weights1 - weights2

    # NOTE: a paper does "len(weights) -> len(weights) * perturbation_std,
    # but it's just a matter of a different tuning for l2_coef
    delta = jax.tree_multimap(
        lambda y: 1. /
        (len(weights)) * jnp.sum(
            y * jnp.reshape(weights, ([weights.shape[0]] + [1] *
                                      (len(y.shape) - 1))),
            axis=0), noise)
    # l2coeff controls the weight decay of the parameters of our policy network.
    # This prevents the parameters from growing very large compared to the
    # perturbations.
    delta = jax.tree_multimap(lambda d, th: d - l2coeff * th, delta,
                              training_state.policy_params)
    delta = jax.tree_map(lambda x: -x, delta)

    params_update, optimizer_state = optimizer.update(
        delta, training_state.optimizer_state)
    policy_params = optax.apply_updates(training_state.policy_params,
                                        params_update)

    metrics = {
        'params_norm': optax.global_norm(policy_params),
        'eval_scores_mean': jnp.mean(eval_scores),
        'eval_scores_std': jnp.std(eval_scores),
        'weights': jnp.mean(weights),
    }
    return (state,
            TrainingState(
                key=key,
                normalizer_params=normalizer_params,
                optimizer_state=optimizer_state,
                policy_params=policy_params), metrics)

  training_state = TrainingState(
      key=key,
      normalizer_params=normalizer_params,
      optimizer_state=optimizer_state,
      policy_params=policy_params)

  training_walltime = 0
  eval_walltime = 0
  sps = 0
  eval_sps = 0
  metrics = {}
  summary = {}
  state = first_state
  it = -1

  while True:
    it += 1
    logging.info('starting iteration %s %s', it, time.time() - xt)
    t = time.time()
    num_process_env_steps = int(
        training_state.normalizer_params[0]) * action_repeat

    if process_id == 0 and it % log_frequency == 0:
      eval_state, key_debug = (
          run_eval(eval_first_state, key_debug,
                   training_state.policy_params,
                   training_state.normalizer_params))
      eval_state.completed_episodes.block_until_ready()
      eval_walltime += time.time() - t
      eval_sps = (episode_length * eval_first_state.core.reward.shape[0] /
                  (time.time() - t))
      avg_episode_length = (
          eval_state.completed_episodes_steps / eval_state.completed_episodes)
      metrics = dict(
          dict({
              f'eval/episode_{name}': value / eval_state.completed_episodes
              for name, value in eval_state.completed_episodes_metrics.items()
          }),
          **dict({
              'eval/completed_episodes': eval_state.completed_episodes,
              'eval/avg_episode_length': avg_episode_length,
              'speed/sps': sps,
              'speed/eval_sps': eval_sps,
              'speed/training_walltime': training_walltime,
              'speed/eval_walltime': eval_walltime,
              'speed/timestamp': training_walltime,
              'train/completed_episodes': it * num_envs,
              'train/params_norm': summary.get('params_norm', 0),
              'train/eval_scores_mean': summary.get('eval_scores_mean', 0),
              'train/eval_scores_std': summary.get('eval_scores_std', 0),
              'train/weights': summary.get('weights', 0),
          }))
      logging.info('Step %s metrics %s', num_process_env_steps, metrics)
      if progress_fn:
        progress_fn(num_process_env_steps, metrics)

    if num_process_env_steps > num_timesteps:
      break

    t = time.time()
    # optimization
    new_state, training_state, summary = es_one_epoch(state, training_state)
    del new_state
    # Don't override state with new_state. For environments with variable
    # episode length we still want to start from a 'reset', not from where the
    # last run finished.
    jax.tree_map(lambda x: x.block_until_ready(), training_state)
    sps = (int(training_state.normalizer_params[0]) * action_repeat -
           num_process_env_steps) / (
               time.time() - t)
    training_walltime += time.time() - t

  _, inference = make_params_and_inference_fn(core_env.observation_size,
                                              core_env.action_size,
                                              normalize_observations)
  params = training_state.normalizer_params, training_state.policy_params

  return (inference, params, metrics)


def make_es_model(parametric_action_distribution, obs_size):
  return networks.make_model(
      [32, 32, 32, 32, parametric_action_distribution.param_size], obs_size)


def make_params_and_inference_fn(observation_size, action_size,
                                 normalize_observations):
  """Creates params and inference function for the ES agent."""
  obs_normalizer_params, obs_normalizer_apply_fn = normalization.make_data_and_apply_fn(
      observation_size, normalize_observations)
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size)
  policy_model = make_es_model(parametric_action_distribution,
                               observation_size)

  def inference_fn(params, obs, key):
    normalizer_params, policy_params = params
    obs = obs_normalizer_apply_fn(normalizer_params, obs)
    action = parametric_action_distribution.sample(
        policy_model.apply(policy_params, obs), key)
    return action

  params = (obs_normalizer_params, policy_model.init(jax.random.PRNGKey(0)))
  return params, inference_fn



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Just some tree functions."""

from typing import List, Optional

import dataclasses
import brax

@dataclasses.dataclass
class Node(object):
  """A node within the kinematic tree of a system."""
  name: str = ''
  parent: Optional['Node'] = None
  children: List['Node'] = dataclasses.field(default_factory=list)

  def add(self, node: 'Node'):
    if node.parent:
      node.parent.children.remove(node)
    node.parent = self
    self.children.append(node)

  def find(self, name: str) -> Optional['Node']:
    """Recursively finds a node with parent with this name."""
    if self.name == name:
      return self
    for c in self.children:
      v = c.find(name)
      if v:
        return v
    return None

  def depth_first(self):
    """Iterates through a tree depth first."""
    for c in self.children:
      yield c
      yield from c.depth_first()

  @classmethod
  def from_config(cls, config: brax.Config) -> 'Node':
    root = Node()

    for joint in config.joints:
      parent = root.find(joint.parent)
      child = root.find(joint.child)
      if not parent:
        parent = Node(name=joint.parent)
        root.add(parent)
      if not child:
        child = Node(name=joint.child)
      parent.add(child)

    for body in config.bodies:
      if not root.find(body.name):
        root.add(Node(name=body.name))

    return root



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers for Brax and Gym env."""

from typing import ClassVar, Optional

from brax.envs import env as brax_env
import gym
from gym import spaces
from gym.vector import utils
import jax
from jax import numpy as jnp
import numpy as np


class VectorWrapper(brax_env.Wrapper):
  """Vectorizes Brax env."""

  def __init__(self, env: brax_env.Env, batch_size: int):
    super().__init__(env)
    self.batch_size = batch_size

  def reset(self, rng: jnp.ndarray) -> brax_env.State:
    rng = jax.random.split(rng, self.batch_size)
    return jax.vmap(self.env.reset)(rng)

  def step(self, state: brax_env.State, action: jnp.ndarray) -> brax_env.State:
    return jax.vmap(self.env.step)(state, action)


class EpisodeWrapper(brax_env.Wrapper):
  """Maintains episode step count and sets done at episode end."""

  def __init__(self, env: brax_env.Env, episode_length: int,
               action_repeat: int):
    super().__init__(env)
    if hasattr(self.unwrapped, 'sys'):
      self.unwrapped.sys.config.dt *= action_repeat
      self.unwrapped.sys.config.substeps *= action_repeat
    self.episode_length = episode_length
    self.action_repeat = action_repeat

  def reset(self, rng: jnp.ndarray) -> brax_env.State:
    state = self.env.reset(rng)
    state.info['steps'] = jnp.zeros(())
    return state

  def step(self, state: brax_env.State, action: jnp.ndarray) -> brax_env.State:
    state = self.env.step(state, action)
    steps = state.info['steps'] + self.action_repeat
    one = jnp.ones_like(state.done)
    done = jnp.where(steps >= self.episode_length, one, state.done)
    state.info['steps'] = steps
    return state.replace(done=done)


class AutoResetWrapper(brax_env.Wrapper):
  """Automatically resets Brax envs that are done."""

  def reset(self, rng: jnp.ndarray) -> brax_env.State:
    state = self.env.reset(rng)
    state.info['first_qp'] = state.qp
    state.info['first_obs'] = state.obs
    return state

  def step(self, state: brax_env.State, action: jnp.ndarray) -> brax_env.State:
    steps = state.info['steps']
    steps = jnp.where(state.done, jnp.zeros_like(steps), steps)
    state.info.update(steps=steps)
    state = state.replace(done=jnp.zeros_like(state.done))
    state = self.env.step(state, action)

    def where_done(x, y):
      done = state.done
      if done.shape:
        done = jnp.reshape(done, [x.shape[0]] + [1] * (len(x.shape) - 1))  # type: ignore
      return jnp.where(done, x, y)

    qp = jax.tree_map(where_done, state.info['first_qp'], state.qp)
    obs = where_done(state.info['first_obs'], state.obs)
    return state.replace(qp=qp, obs=obs)


class GymWrapper(gym.Env):
  """A wrapper that converts Brax Env to one that follows Gym API."""

  # Flag that prevents `gym.register` from misinterpreting the `_step` and
  # `_reset` as signs of a deprecated gym Env API.
  _gym_disable_underscore_compat: ClassVar[bool] = True

  def __init__(self,
               env: brax_env.Env,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.seed(seed)
    self.backend = backend
    self._state = None

    obs_high = (np.inf * np.ones(self._env.observation_size)).astype(
        np.float32)
    self.observation_space = spaces.Box(-obs_high, obs_high, dtype=np.float32)

    action_high = np.ones(self._env.action_size, dtype=np.float32)
    self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)

    def reset(key):
      key1, key2 = jax.random.split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      return state, state.obs, state.reward, state.done

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    return obs

  def step(self, action):
    self._state, obs, reward, done = self._step(self._state, action)
    return obs, reward, done, {}

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)


class VectorGymWrapper(gym.vector.VectorEnv):
  """A wrapper that converts batched Brax Env to one that follows Gym VectorEnv API."""

  # Flag that prevents `gym.register` from misinterpreting the `_step` and
  # `_reset` as signs of a deprecated gym Env API.
  _gym_disable_underscore_compat: ClassVar[bool] = True

  def __init__(self,
               env: brax_env.Env,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    if not hasattr(self._env, 'batch_size'):
      raise ValueError('underlying env must be batched')

    self.num_envs = self._env.batch_size
    self.seed(seed)
    self.backend = backend
    self._state = None

    obs_high = (np.inf * np.ones(self._env.observation_size)).astype(np.float32)
    self.single_observation_space = spaces.Box(
        -obs_high, obs_high, dtype=np.float32)
    self.observation_space = utils.batch_space(self.single_observation_space,
                                               self.num_envs)

    action_high = np.ones(self._env.action_size, dtype=np.float32)
    self.single_action_space = spaces.Box(
        -action_high, action_high, dtype=np.float32)
    self.action_space = utils.batch_space(self.single_action_space,
                                          self.num_envs)

    def reset(key):
      key1, key2 = jax.random.split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      return state, state.obs, state.reward, state.done

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    return obs

  def step(self, action):
    self._state, obs, reward, done = self._step(self._state, action)
    return obs, reward, done, {}

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Analytic policy gradient tests."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax.training import apg


class APGTest(parameterized.TestCase):
  """Tests for APG module."""

  def testTraining(self):
    _, _, metrics = apg.train(
        environment_fn=envs.create_fn('reacherangle'),
        episode_length=100,
        action_repeat=4,
        num_envs=16,
        learning_rate=3e-3,
        normalize_observations=True,
        log_frequency=200,
        truncation_length=10,
    )
    self.assertGreater(metrics['eval/episode_reward'], -2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a reacher to reach a target.

Based on the OpenAI Gym MuJoCo Reacher environment.
"""

from typing import Tuple

import brax
from brax.envs import env
from brax.physics import math
from brax.physics.base import take
import jax
import jax.numpy as jnp


class Reacher(env.Env):
  """Trains a reacher arm to touch a sequence of random targets."""

  def __init__(self, **kwargs):
    super().__init__(_SYSTEM_CONFIG, **kwargs)
    self.target_idx = self.sys.body_idx['target']
    self.arm_idx = self.sys.body_idx['body1']

  def reset(self, rng: jnp.ndarray) -> env.State:
    qp = self.sys.default_qp()
    _, target = self._random_target(rng)
    pos = jax.ops.index_update(qp.pos, jax.ops.index[self.target_idx], target)
    qp = qp.replace(pos=pos)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jnp.zeros(3)
    metrics = {
        'rewardDist': zero,
        'rewardCtrl': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jnp.ndarray) -> env.State:
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    # vector from tip to target is last 3 entries of obs vector
    reward_dist = -jnp.linalg.norm(obs[-3:])
    reward_ctrl = -jnp.square(action).sum()
    reward = reward_dist + reward_ctrl

    state.metrics.update(
        rewardDist=reward_dist,
        rewardCtrl=reward_ctrl,
    )

    return state.replace(qp=qp, obs=obs, reward=reward)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jnp.ndarray:
    """Egocentric observation of target and arm body."""

    # some pre-processing to pull joint angles and velocities
    (joint_angle,), _ = self.sys.joint_revolute.angle_vel(qp)

    # qpos:
    # x,y coord of target
    qpos = [qp.pos[self.target_idx, :2]]

    # dist to target and speed of tip
    arm_qps = take(qp, jnp.array(self.arm_idx))
    tip_pos, tip_vel = math.to_world(arm_qps, jnp.array([0.11, 0., 0.]))
    tip_to_target = [tip_pos - qp.pos[self.target_idx]]
    cos_sin_angle = [jnp.cos(joint_angle), jnp.sin(joint_angle)]

    # qvel:
    # velocity of tip
    qvel = [tip_vel[:2]]

    return jnp.concatenate(cos_sin_angle + qpos + qvel + tip_to_target)

  def _random_target(self, rng: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Returns a target location in a random circle slightly above xy plane."""
    rng, rng1, rng2 = jax.random.split(rng, 3)
    dist = .2 * jax.random.uniform(rng1)
    ang = jnp.pi * 2. * jax.random.uniform(rng2)
    target_x = dist * jnp.cos(ang)
    target_y = dist * jnp.sin(ang)
    target_z = .01
    target = jnp.array([target_x, target_y, target_z]).transpose()
    return rng, target


_SYSTEM_CONFIG = """
bodies {
  name: "ground"
  colliders {
    plane {
    }
  }
  mass: 1.0
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  frozen {
    all: true
  }
}
bodies {
  name: "body0"
  colliders {
    position {
      x: 0.05
    }
    rotation {
      y: 90.0
    }
    capsule {
      radius: 0.01
      length: 0.12
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 0.035604715
}
bodies {
  name: "body1"
  colliders {
    position {
      x: 0.05
    }
    rotation {
      y: 90.0
    }
    capsule {
      radius: 0.01
      length: 0.12
    }
  }
  colliders {
    position { x: .11 }
    sphere {
      radius: 0.01
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 0.035604715
}
bodies {
  name: "target"
  colliders {
    position {
    }
    sphere {
      radius: 0.009
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen { all: true }
}
joints {
  name: "joint0"
  stiffness: 100.0
  parent: "ground"
  child: "body0"
  parent_offset {
    z: 0.01
  }
  child_offset {
  }
  rotation {
    y: -90.0
  }
  angle_limit {
      min: -360
      max: 360
    }
  limit_strength: 0.0
  spring_damping: 3.0
}
joints {
  name: "joint1"
  stiffness: 100.0
  parent: "body0"
  child: "body1"
  parent_offset {
    x: 0.1
  }
  child_offset {
  }
  rotation {
    y: -90.0
  }
  angle_limit {
    min: -360
    max: 360
  }
  limit_strength: 0.0
  spring_damping: 3.0
}
actuators {
  name: "joint0"
  joint: "joint0"
  strength: 25.0
  torque {
  }
}
actuators {
  name: "joint1"
  joint: "joint1"
  strength: 25.0
  torque {
  }
}
collide_include {
}
friction: 0.6
gravity {
  z: -9.81
}
baumgarte_erp: 0.1
dt: 0.02
substeps: 4
frozen {
  position {
    z: 1.0
  }
  rotation {
    x: 1.0
    y: 1.0
  }
}
"""



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""A collection of integrators."""

import jax
import jax.numpy as jnp
from brax.physics import config_pb2
from brax.physics import math
from brax.physics.base import P, QP, vec_to_np


def kinetic(_, qp: QP, dt: float, active_pos: jnp.ndarray,
            active_rot: jnp.ndarray) -> QP:
  """Performs a kinetic integration step.

  Args:
    qp: State data to be integrated
    dt: Timestep length
    active_pos: [# bodies, 3] mask for active translational dofs
    active_rot: [# bodes, 3] mask for active rotational dofs

  Returns:
    State data advanced by one kinematic integration step.
  """

  @jax.vmap
  def op(qp: QP, active_pos: jnp.ndarray,
         active_rot: jnp.ndarray) -> QP:
    pos = qp.pos + qp.vel * dt * active_pos
    rot_at_ang_quat = math.ang_to_quat(qp.ang * active_rot)
    rot = jnp.matmul(
        jnp.matmul(jnp.eye(4) + .5 * dt * rot_at_ang_quat, qp.rot),
        jnp.eye(4) - .5 * dt * rot_at_ang_quat)
    rot = rot / jnp.linalg.norm(rot)
    return QP(pos=pos, rot=rot, vel=qp.vel, ang=qp.ang)

  return op(qp, active_pos, active_rot)


def potential(config: config_pb2.Config, qp: QP, dp: P, dt: float,
              active_pos: jnp.ndarray, active_rot: jnp.ndarray) -> QP:
  """Performs a potential integration step.

  Args:
    config: Brax system configuration
    qp: State data to be integrated
    dp: Impulses to apply during this potential step
    dt: Timestep length
    active_pos: [# bodies, 3] mask for active translational dofs
    active_rot: [# bodes, 3] mask for active rotational dofs

  Returns:
    State data advanced by one potential integration step.
  """

  @jax.vmap
  def op(qp: QP, dp: P, active_pos: jnp.ndarray,
         active_rot: jnp.ndarray) -> QP:
    vel = (jnp.exp(config.velocity_damping * dt) * qp.vel +
           (dp.vel + vec_to_np(config.gravity)) * dt) * active_pos
    ang = (jnp.exp(config.angular_damping * dt) * qp.ang +
           dp.ang * dt) * active_rot
    return QP(pos=qp.pos, rot=qp.rot, vel=vel, ang=ang)

  return op(qp, dp, active_pos, active_rot)


def potential_collision(_, qp: QP, dp: P, active_pos: jnp.ndarray,
                        active_rot: jnp.ndarray) -> QP:
  """Performs a potential collision integration step.

  Args:
    qp: State data to be integrated
    dp: Velocity-level collision updates to apply this integration step
    active_pos: [# bodies, 3] mask for active translational dofs
    active_rot: [# bodes, 3] mask for active rotational dofs

  Returns:
    State data advanced by one velocity-level update."""

  @jax.vmap
  def op(qp: QP, dp: P, active_pos: jnp.ndarray,
         active_rot: jnp.ndarray) -> QP:
    vel = (qp.vel + dp.vel) * active_pos
    ang = (qp.ang + dp.ang) * active_rot
    return QP(pos=qp.pos, rot=qp.rot, vel=vel, ang=ang)

  return op(qp, dp, active_pos, active_rot)



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core brax structs and some conversion and slicing functions."""

from flax import struct
import jax
import jax.numpy as jnp
from brax.physics import config_pb2


@struct.dataclass
class Q(object):
  """Coordinates: position and rotation.

  Attributes:
    pos: Location of center of mass.
    rot: Rotation about center of mass, represented as a quaternion.
  """
  pos: jnp.ndarray
  rot: jnp.ndarray

  def __add__(self, o):
    if isinstance(o, P):
      return QP(self.pos, self.rot, o.vel, o.ang)
    elif isinstance(o, Q):
      return Q(self.pos + o.pos, self.rot + o.rot)
    elif isinstance(o, QP):
      return QP(self.pos + o.pos, self.rot + o.rot, o.vel, o.ang)
    else:
      raise ValueError("add only supported for P, Q, QP")


@struct.dataclass
class P(object):
  """Time derivatives: velocity and angular velocity.

  Attributes:
    vel: Velocity.
    ang: Angular velocity about center of mass.
  """
  vel: jnp.ndarray
  ang: jnp.ndarray

  def __add__(self, o):
    if isinstance(o, P):
      return P(self.vel + o.vel, self.ang + o.ang)
    elif isinstance(o, Q):
      return QP(o.pos, o.rot, self.vel, self.ang)
    elif isinstance(o, QP):
      return QP(o.pos, o.rot, self.vel + o.vel, self.ang + o.ang)
    else:
      raise ValueError("add only supported for P, Q, QP")

  def __mul__(self, o):
    return P(self.vel * o, self.ang * o)


@struct.dataclass
class QP(object):
  """A coordinate and time derivative frame for a brax body.

  Attributes:
    pos: Location of center of mass.
    rot: Rotation about center of mass, represented as a quaternion.
    vel: Velocity.
    ang: Angular velocity about center of mass.
  """
  pos: jnp.ndarray
  rot: jnp.ndarray
  vel: jnp.ndarray
  ang: jnp.ndarray

  def __add__(self, o):
    if isinstance(o, P):
      return QP(self.pos, self.rot, self.vel + o.vel, self.ang + o.ang)
    elif isinstance(o, Q):
      return QP(self.pos + o.pos, self.rot + o.rot, self.vel, self.ang)
    elif isinstance(o, QP):
      return QP(self.pos + o.pos, self.rot + o.rot, self.vel + o.vel,
                self.ang + o.ang)
    else:
      raise ValueError("add only supported for P, Q, QP")

  def __mul__(self, o):
    return QP(self.pos * o, self.rot * o, self.vel * o, self.ang * o)

  @classmethod
  def zero(cls):
    return cls(
        pos=jnp.zeros(3),
        rot=jnp.array([1., 0., 0., 0]),
        vel=jnp.zeros(3),
        ang=jnp.zeros(3))


@struct.dataclass
class Info(object):
  """Auxilliary data calculated during the dynamics of each physics step.

  Attributes:
    contact: External contact forces applied at a step
    joint: Joint constraint forces applied at a step
    actuator: Actuator forces applied at a step
  """
  contact: P
  joint: P
  actuator: P


def vec_to_np(v):
  return jnp.array([v.x, v.y, v.z])


def quat_to_np(q):
  return jnp.array([q.w, q.x, q.y, q.z])


def euler_to_quat(v):
  """Converts euler rotations in degrees to quaternion."""
  # this follows the Tait-Bryan intrinsic rotation formalism: x-y'-z''
  c1, c2, c3 = jnp.cos(jnp.array([v.x, v.y, v.z]) * jnp.pi / 360)
  s1, s2, s3 = jnp.sin(jnp.array([v.x, v.y, v.z]) * jnp.pi / 360)
  w = c1 * c2 * c3 - s1 * s2 * s3
  x = s1 * c2 * c3 + c1 * s2 * s3
  y = c1 * s2 * c3 - s1 * c2 * s3
  z = c1 * c2 * s3 + s1 * s2 * c3
  return jnp.array([w, x, y, z])


def take(objects, i: jnp.ndarray, axis=0):
  """Returns objects sliced by i."""
  flat_data, py_tree_def = jax.tree_flatten(objects)
  sliced_data = [jnp.take(k, i, axis=axis, mode="clip") for k in flat_data]
  return jax.tree_unflatten(py_tree_def, sliced_data)


def validate_config(config: config_pb2.Config) -> config_pb2.Config:
  """Validate and normalize config settings for use in systems."""
  if config.dt <= 0:
    raise RuntimeError("config.dt must be positive")

  def find_dupes(objs):
    names = set()
    for obj in objs:
      if obj.name in names:
        raise RuntimeError(f"duplicate name in config: {obj.name}")
      names.add(obj.name)

  find_dupes(config.bodies)
  find_dupes(config.joints)
  find_dupes(config.actuators)

  # TODO: more config validation

  # reify all frozen dimensions in the system
  allvec = config_pb2.Vector3(x=1.0, y=1.0, z=1.0)
  frozen = config.frozen
  if frozen.all:
    frozen.position.CopyFrom(allvec)
    frozen.rotation.CopyFrom(allvec)
  if all([
      frozen.position.x, frozen.position.y, frozen.position.z,
      frozen.rotation.x, frozen.rotation.y, frozen.rotation.z
  ]):
    config.frozen.all = True
  for b in config.bodies:
    b.frozen.position.x = b.frozen.position.x or frozen.position.x
    b.frozen.position.y = b.frozen.position.y or frozen.position.y
    b.frozen.position.z = b.frozen.position.z or frozen.position.z
    b.frozen.rotation.x = b.frozen.rotation.x or frozen.rotation.x
    b.frozen.rotation.y = b.frozen.rotation.y or frozen.rotation.y
    b.frozen.rotation.z = b.frozen.rotation.z or frozen.rotation.z
    if b.frozen.all:
      b.frozen.position.CopyFrom(allvec)
      b.frozen.rotation.CopyFrom(allvec)
    if all([
        b.frozen.position.x, b.frozen.position.y, b.frozen.position.z,
        b.frozen.rotation.x, b.frozen.rotation.y, b.frozen.rotation.z
    ]):
      b.frozen.all = True
  frozen.all = all(b.frozen.all for b in config.bodies)

  return config



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.physics."""

from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax import numpy as jnp
import brax
from brax.physics.base import take
from google.protobuf import text_format


class BodyTest(absltest.TestCase):

  def test_projectile_motion(self):
    """A ball with an initial velocity curves down due to gravity."""
    sys = brax.System(
        text_format.Parse(
            """
    dt: 1 substeps: 1000
    gravity { z: -9.8 }
    bodies { name: "Ball" mass: 1 }
    defaults { qps { name: "Ball" vel {x: 1}}}
    """, brax.Config()))
    qp = sys.default_qp()
    qp, _ = sys.step(qp, jnp.array([]))
    # v = v_0 + a * t
    self.assertAlmostEqual(qp.vel[0, 2], -9.8, 2)
    # x = x_0 + v_0 * t + 0.5 * a * t^2
    self.assertAlmostEqual(qp.pos[0, 0], 1, 2)
    self.assertAlmostEqual(qp.pos[0, 2], -9.8 / 2, 2)


class BoxTest(absltest.TestCase):

  _CONFIG = """
    dt: 1.5 substeps: 1000 friction: 0.6 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.5 y: 0.5 z: 0.5 }}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults { qps { name: "box" pos { z: 1 }}}
    defaults { qps { name: "box" pos { z: 2 } vel {x: 2}}}
  """

  def test_box_hits_ground(self):
    """A box falls onto the ground and stops."""
    sys = brax.System(text_format.Parse(BoxTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, _ = sys.step(qp, jnp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)

  def test_box_slide(self):
    """A box slides across the ground and comes to a stop."""
    sys = brax.System(text_format.Parse(BoxTest._CONFIG, brax.Config()))
    qp = sys.default_qp(1)
    qp, _ = sys.step(qp, jnp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)
    self.assertGreater(qp.pos[0, 0], 1)  # after sliding for a bit...
    self.assertAlmostEqual(qp.vel[0, 0], 0, 2)  # friction brings it to a stop
    self.assertLess(qp.pos[0, 0], 1.5)  # ... and keeps it from travelling 2m


class HeightMapTest(absltest.TestCase):

  _CONFIG = """
    dt: 2 substeps: 1000 friction: 1 baumgarte_erp: 0.1 elasticity: 0
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.3 y: 0.3 z: 0.3 }}}
      inertia { x: 0.1 y: 0.1 z: 0.1 }
    }
    bodies {
      name: "ground"
      frozen: { all: true }
      colliders {
        heightMap {
          size: 10
          data: [0, 0, 0, 0, 0, 0, 0, 0, 0]
        }
      }
    }
    defaults { qps { name: "box" pos: {x: 5 y: 5 z: 1}}}
  """

  def test_box_stays_on_heightMap (self):
    """A box falls onto the height map and stops."""
    sys = brax.System(text_format.Parse(HeightMapTest._CONFIG, brax.Config()))
    qp = sys.default_qp()
    qp, _ = sys.step(qp, [])
    self.assertAlmostEqual(qp.pos[0, 2], 0.3, 2)


class SphereTest(absltest.TestCase):

  _CONFIG = """
    dt: 5 substeps: 5000 friction: 0.6 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "Sphere1" mass: 1
      colliders { sphere { radius: 0.25}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {qps { name: "Sphere1" pos {z: 1}}}
  """

  def test_sphere_hits_ground(self):
    """A sphere falls onto the ground and stops."""
    sys = brax.System(text_format.Parse(SphereTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, _ = sys.step(qp, jnp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.25, 2)


class CapsuleTest(absltest.TestCase):

  _CONFIG = """
    dt: 5 substeps: 5000 friction: 0.6 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "Capsule1" mass: 1
      colliders { capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule2" mass: 1
      colliders { rotation { y: 90 } capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule3" mass: 1
      colliders { rotation { y: 45 } capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule4" mass: 1
      colliders { rotation { x: 45 } capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {
      qps { name: "Capsule1" pos {z: 1}}
      qps { name: "Capsule2" pos {x:1 z: 1}}
      qps { name: "Capsule3" pos {x:3 z: 1}}
      qps { name: "Capsule4" pos {x:5 z: 1}}
    }
    defaults {
      qps { name: "Capsule1" pos {z: 1}}
      qps { name: "Capsule2" pos {z: 2}}
      qps { name: "Capsule3" pos {x:3 z: 1}}
      qps { name: "Capsule4" pos {x:5 z: 1}}
    }
  """

  def test_capsule_hits_ground(self):
    """A capsule falls onto the ground and stops."""
    sys = brax.System(text_format.Parse(CapsuleTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, _ = sys.step(qp, jnp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)  # standing up and down
    self.assertAlmostEqual(qp.pos[1, 2], 0.25, 2)  # lying on its side
    self.assertAlmostEqual(qp.pos[2, 2], 0.25, 2)  # rolls to side from y rot
    self.assertAlmostEqual(qp.pos[3, 2], 0.25, 2)  # rolls to side from x rot

  def test_capsule_hits_capsule(self):
    """A capsule falls onto another capsule and balances on it."""
    sys = brax.System(text_format.Parse(CapsuleTest._CONFIG, brax.Config()))
    qp = sys.default_qp(1)
    qp, _ = sys.step(qp, jnp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)  # standing up and down
    self.assertAlmostEqual(qp.pos[1, 2], 1.25, 2)  # lying on Capsule1


class JointTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 100000
    gravity { z: -9.8 }
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Bob" }
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit { min: -180 max: 180 }
    }
  """

  @parameterized.parameters((2.0, 0.125, 0.0625), (5.0, 0.125, 0.03125),
                            (1.0, 0.0625, 0.1))
  def test_pendulum_period(self, mass, radius, vel):
    """A small spherical mass swings for approximately one period."""
    config = text_format.Parse(JointTest._CONFIG, brax.Config())

    # this length of time comes from the following:
    # inertia_about_anchor = mass * dist_to_anchor^2 + inertia_cm
    # dist_to_anchor = 1.0
    # inertia_cm = 2/5 * mass * radius^2 (solid sphere)
    # T = 2 * pi * sqrt(inertia_about_anchor / (2 * mass * g * dist_to_anchor))
    config.dt = float(2 * jnp.pi * jnp.sqrt((.4 * radius**2 + 1.) / 9.8))
    config.bodies[1].mass = mass
    config.bodies[1].inertia.x = .4 * mass * radius**2
    config.bodies[1].inertia.y = .4 * mass * radius**2
    config.bodies[1].inertia.z = .4 * mass * radius**2
    sys = brax.System(config)

    # initializing system to have a small initial velocity and ang velocity
    # so that small angle approximation is valid
    qp = brax.QP(
        pos=jnp.array([[0., 0., -1.], [0., 0., 0.]]),
        rot=jnp.array([[1., 0., 0., 0.], [1., 0., 0., 0.]]),
        vel=jnp.array([[0., vel, 0.], [0., 0., 0.]]),
        ang=jnp.array([[.5 * vel, 0., 0.], [0., 0., 0.]]))
    qp, _ = sys.step(qp, jnp.array([]))
    self.assertAlmostEqual(qp.pos[0, 1], 0., 3)  # returned to the origin


class Actuator1DTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 1200
    dt: 4.0
    gravity { z: -9.8 }
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }}
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit { min: -180 max: 180 }
      angular_damping: 140.0
    }
    actuators {
      name: "Joint"
      joint: "Joint"
      strength: 15000.0
      angle {}
    }
    defaults { qps { name: "Anchor" pos {z: 2}} qps { name: "Bob" pos {z: 1}}}
    defaults { angles { name: "Joint" angle: { x: 60 } } }
  """

  @parameterized.parameters(15., 30., 45., 90.)
  def test_1d_angle_actuator(self, target_angle):
    """A simple part actuates to a target angle."""
    config = text_format.Parse(Actuator1DTest._CONFIG, brax.Config())
    sys = brax.System(config=config)
    qp = sys.default_qp()
    qp, _ = sys.step(qp, jnp.array([target_angle]))
    qp_p = take(qp, 0)
    qp_c = take(qp, 1)
    joint = take(sys.joint_revolute, 0)
    _, (angle,) = joint.axis_angle(qp_p, qp_c)

    self.assertAlmostEqual(target_angle * jnp.pi / 180, angle, 2)


class Actuator2DTest(parameterized.TestCase):

  _CONFIG = """
  substeps: 2000
    dt: 2.0
    gravity { z: -9.8 }
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }
      colliders { capsule { radius: 0.5 length: 2.0 }}
    }
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit {
        min: -180
        max: 180
      }
      angle_limit {
        min: -180
        max: 180
      }
      angular_damping: 200.0
    }
    actuators {
      name: "Joint"
      joint: "Joint"
      strength: 2000.0
      angle {}
    }
    defaults { qps { name: "Anchor" pos {z: 2}} qps { name: "Bob" pos {z: 1}}}
  """

  @parameterized.parameters((15., 30.), (45., 90.5), (-120, 60.), (30., -120.),
                            (-150, -130), (130, 165))
  def test_2d_angle_actuator(self, target_angle_1, target_angle_2):
    """A simple part actuates 2d-angle actuator to two target angles."""
    config = text_format.Parse(Actuator2DTest._CONFIG, brax.Config())
    sys = brax.System(config=config)
    qp = sys.default_qp()
    qp, _ = sys.step(qp, jnp.array([target_angle_1, target_angle_2]))
    qp_p = take(qp, 0)
    qp_c = take(qp, 1)
    joint = take(sys.joint_universal, 0)
    _, angles = joint.axis_angle(qp_p, qp_c)

    self.assertAlmostEqual(target_angle_1 * jnp.pi / 180, angles[0], 2)
    self.assertAlmostEqual(target_angle_2 * jnp.pi / 180, angles[1], 2)


class Actuator3DTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 8
    dt: .02
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }
      colliders { capsule { radius: 0.5 length: 2.0 } }
    }
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit {
        min: -100
        max: 100
      }
      angle_limit {
        min: -100
        max: 100
      }
      angle_limit {
        min: -100
        max: 100
      }
      angular_damping: 120.0
    }
    actuators {
      name: "Joint"
      joint: "Joint"
      strength: 40.0
      torque {}
    }
    defaults { qps { name: "Anchor" pos {z: 2}} qps { name: "Bob" pos {z: 1}}}
  """
  torques = [(1, 0, 0), (0, 1, 0), (0, 0, 1), (1, 1, 1)]

  @parameterized.parameters(((15, 15, 15), torques), ((35, 40, 75), torques),
                            ((80, 45, 30), torques))
  def test_3d_torque_actuator(self, limits, torques):
    """A simple part actuates 3d-torque actuator to its limits."""
    config = text_format.Parse(Actuator3DTest._CONFIG, brax.Config())
    for t in torques:
      for angle_limit, limit in zip(config.joints[0].angle_limit, limits):
        angle_limit.min = -limit
        angle_limit.max = limit

      sys = brax.System(config=config)
      qp = sys.default_qp()

      # cuts down compile time for test to only compile a short step
      for _ in range(1000):
        qp, _ = sys.step(qp, jnp.array(t))

      qp_p = take(qp, 0)
      qp_c = take(qp, 1)
      joint = take(sys.joint_spherical, 0)
      _, angles = joint.axis_angle(qp_p, qp_c)
      angles = [a * 180 / jnp.pi for a in angles]
      for angle, limit, torque in zip(angles, limits, t):
        if torque != 0:
          self.assertAlmostEqual(angle, limit, 1)  # actuated to target angle


if __name__ == '__main__':
  absltest.main()
